import pandas as pd
import numpy as np

from typing import Any
from pandas.core.groupby import DataFrameGroupBy
from enum import Enum
from uuid import uuid4


class WindowFunction(str, Enum):
    """Windowing functions for window verb."""

    RowNumber = "row_number"
    Rank = "rank"
    PercentRank = "percent_rank"
    CumulativeDistribution = "cume_dist"
    FirstValue = "first_value"
    LastValue = "last_value"
    FillDown = "fill_down"
    FillUp = "fill_up"
    UUID = "uuid"


def _get_window_indexer(
    column: pd.Series, fixed_size: bool = False
) -> int | pd.api.indexers.BaseIndexer:
    if fixed_size:
        return pd.api.indexers.FixedForwardWindowIndexer(window_size=len(column))

    return len(column)


__window_function_map = {
    WindowFunction.RowNumber: lambda column: column.rolling(
        window=_get_window_indexer(column), min_periods=1
    ).count(),
    WindowFunction.Rank: lambda column: column.rolling(
        window=_get_window_indexer(column), min_periods=1
    ).count(),
    WindowFunction.PercentRank: lambda column: (
        column.rolling(window=_get_window_indexer(column), min_periods=1).count() - 1
    )
    / (len(column) - 1),
    WindowFunction.CumulativeDistribution: lambda column: column.rolling(
        window=_get_window_indexer(column), min_periods=1
    ).count()
    / len(column),
    WindowFunction.FirstValue: lambda column: column.rolling(
        window=_get_window_indexer(column), min_periods=1
    ).apply(lambda x: x.iloc[0]),
    WindowFunction.LastValue: lambda column: column.rolling(
        window=_get_window_indexer(column, True),
        min_periods=1,
    ).apply(lambda x: x.iloc[-1]),
    WindowFunction.FillDown: lambda column: column.rolling(
        window=len(column), min_periods=1
    ).apply(lambda x: x.dropna().iloc[-1]),
    WindowFunction.FillUp: lambda column: column.rolling(
        window=_get_window_indexer(column, True),
        min_periods=1,
    ).apply(lambda x: x.dropna().iloc[0] if np.isnan(x.iloc[0]) else x.iloc[0]),
    WindowFunction.UUID: lambda column: column.apply(lambda _x: str(uuid4())),
}


def window(
    table: pd.DataFrame,
    column: str,
    to: str,
    operation: str,
    **_kwargs: Any,
) -> pd.DataFrame:
    """Apply a window function to a column in a table."""
    window_operation = WindowFunction(operation)
    window = __window_function_map[window_operation](table[column])

    if isinstance(table, DataFrameGroupBy):
        # ungroup table to add new column
        output = table.obj
        output[to] = window.reset_index()[column]
        # group again by original group by
        output = output.groupby(table.keys)
    else:
        output = table
        output[to] = window

    return output



import pandas as pd

from dataclasses import dataclass
from typing import Any, cast
from enum import Enum


@dataclass
class Aggregation:
    """Aggregation class method definition."""

    column: str | None
    operation: str
    to: str

    # Only useful for the concat operation
    separator: str | None = None


def _load_aggregations(
    aggregations: list[dict[str, Any]],
) -> dict[str, Aggregation]:
    return {
        aggregation["column"]: Aggregation(
            aggregation["column"], aggregation["operation"], aggregation["to"]
        )
        for aggregation in aggregations
    }


class FieldAggregateOperation(str, Enum):
    """Aggregate operations for fields."""

    Any = "any"
    Count = "count"
    CountDistinct = "distinct"
    Valid = "valid"
    Invalid = "invalid"
    Max = "max"
    Min = "min"
    Sum = "sum"
    Product = "product"
    Mean = "mean"
    Mode = "mode"
    Median = "median"
    StDev = "stdev"
    StDevPopulation = "stdevp"
    Variance = "variance"
    ArrayAgg = "array_agg"
    ArrayAggDistinct = "array_agg_distinct"


_initial_missing = object()


def reduce(function, sequence, initial=_initial_missing):
    """
    reduce(function, iterable[, initial]) -> value

    Apply a function of two arguments cumulatively to the items of a sequence
    or iterable, from left to right, so as to reduce the iterable to a single
    value.  For example, reduce(lambda x, y: x+y, [1, 2, 3, 4, 5]) calculates
    ((((1+2)+3)+4)+5).  If initial is present, it is placed before the items
    of the iterable in the calculation, and serves as a default when the
    iterable is empty.
    """

    it = iter(sequence)

    if initial is _initial_missing:
        try:
            value = next(it)
        except StopIteration:
            raise TypeError(
                "reduce() of empty iterable with no initial value"
            ) from None
    else:
        value = initial

    for element in it:
        value = function(value, element)

    return value


aggregate_operation_mapping = {
    FieldAggregateOperation.Any: "first",
    FieldAggregateOperation.Count: "count",
    FieldAggregateOperation.CountDistinct: "nunique",
    FieldAggregateOperation.Valid: lambda series: series.dropna().count(),
    FieldAggregateOperation.Invalid: lambda series: series.isna().sum(),
    FieldAggregateOperation.Max: "max",
    FieldAggregateOperation.Min: "min",
    FieldAggregateOperation.Sum: "sum",
    FieldAggregateOperation.Product: lambda series: reduce(lambda x, y: x * y, series),
    FieldAggregateOperation.Mean: "mean",
    FieldAggregateOperation.Median: "median",
    FieldAggregateOperation.StDev: "std",
    FieldAggregateOperation.StDevPopulation: "",
    FieldAggregateOperation.Variance: "variance",
    FieldAggregateOperation.ArrayAgg: lambda series: [e for e in series],
    FieldAggregateOperation.ArrayAggDistinct: lambda series: [
        e for e in series.unique()
    ],
}


def _get_pandas_agg_operation(agg: Aggregation) -> Any:
    # TODO: Merge into datashaper
    if agg.operation == "string_concat":
        return (agg.separator or ",").join
    return aggregate_operation_mapping[FieldAggregateOperation(agg.operation)]


def aggregate_override(
    input: pd.DataFrame,
    aggregations: list[dict[str, Any]],
    groupby: list[str] | None = None,
) -> pd.DataFrame:
    """Aggregate method definition."""
    aggregations_to_apply = _load_aggregations(aggregations)
    df_aggregations = {
        agg.column: _get_pandas_agg_operation(agg)
        for agg in aggregations_to_apply.values()
    }
    input_table = input

    if groupby is None:
        output_grouped = input_table.groupby(lambda _x: True)
    else:
        output_grouped = input_table.groupby(groupby, sort=False)
    output = cast(pd.DataFrame, output_grouped.agg(df_aggregations))
    output.rename(
        columns={agg.column: agg.to for agg in aggregations_to_apply.values()},
        inplace=True,
    )
    output.columns = [agg.to for agg in aggregations_to_apply.values()]

    return output.reset_index()



from enum import Enum
from dataclasses import dataclass
import pandas as pd


class SortDirection(str, Enum):
    Ascending = "asc"
    Descending = "desc"


@dataclass
class OrderByInstruction:
    column: str
    direction: SortDirection


def orderby(table: pd.DataFrame, orders: list[dict]) -> pd.DataFrame:
    orders_instructions = [
        OrderByInstruction(
            column=order["column"], direction=SortDirection(order["direction"])
        )
        for order in orders
    ]

    columns = [order.column for order in orders_instructions]
    ascending = [
        order.direction == SortDirection.Ascending for order in orders_instructions
    ]
    return table.sort_values(by=columns, ascending=ascending)



import pandas as pd
from . import schemas


def restore_community_hierarchy(
    input: pd.DataFrame,
    name_column: str = schemas.NODE_NAME,
    community_column: str = schemas.NODE_COMMUNITY,
    level_column: str = schemas.NODE_LEVEL,
    **_kwargs,
) -> pd.DataFrame:
    """Restore the community hierarchy from the node data."""
    node_df: pd.DataFrame = input
    community_df = (
        node_df.groupby([community_column, level_column])
        .agg({name_column: list})
        .reset_index()
    )
    community_levels = {}
    for _, row in community_df.iterrows():
        level = row[level_column]
        name = row[name_column]
        community = row[community_column]

        if community_levels.get(level) is None:
            community_levels[level] = {}
        community_levels[level][community] = name

    # get unique levels, sorted in ascending order
    levels = sorted(community_levels.keys())

    community_hierarchy = []

    for idx in range(len(levels) - 1):
        level = levels[idx]
        # log.debug("Level: %s", level)
        next_level = levels[idx + 1]
        current_level_communities = community_levels[level]
        next_level_communities = community_levels[next_level]
        # log.debug(
        #     "Number of communities at level %s: %s",
        #     level,
        #     len(current_level_communities),
        # )

        for current_community in current_level_communities:
            current_entities = current_level_communities[current_community]

            # loop through next level's communities to find all the subcommunities
            entities_found = 0
            for next_level_community in next_level_communities:
                next_entities = next_level_communities[next_level_community]
                if set(next_entities).issubset(set(current_entities)):
                    community_hierarchy.append(
                        {
                            community_column: current_community,
                            schemas.COMMUNITY_LEVEL: level,
                            schemas.SUB_COMMUNITY: next_level_community,
                            schemas.SUB_COMMUNITY_SIZE: len(next_entities),
                        }
                    )

                    entities_found += len(next_entities)
                    if entities_found == len(current_entities):
                        break

    return pd.DataFrame(community_hierarchy)



import networkx as nx
import pandas as pd
import html

from typing import Any
from utils.graph import load_graph
from utils.hash import gen_uuid
from enum import Enum
from random import Random

from graspologic.partition import hierarchical_leiden
from graspologic.utils import largest_connected_component


def normalize_node_names(graph: nx.Graph | nx.DiGraph) -> nx.Graph | nx.DiGraph:
    """Normalize node names."""
    node_mapping = {node: html.unescape(node.upper().strip()) for node in graph.nodes()}  # type: ignore
    return nx.relabel_nodes(graph, node_mapping)


def _stabilize_graph(graph: nx.Graph) -> nx.Graph:
    """Ensure an undirected graph with the same relationships will always be read the same way."""
    fixed_graph = nx.DiGraph() if graph.is_directed() else nx.Graph()

    sorted_nodes = graph.nodes(data=True)
    sorted_nodes = sorted(sorted_nodes, key=lambda x: x[0])

    fixed_graph.add_nodes_from(sorted_nodes)
    edges = list(graph.edges(data=True))

    # If the graph is undirected, we create the edges in a stable way, so we get the same results
    # for example:
    # A -> B
    # in graph theory is the same as
    # B -> A
    # in an undirected graph
    # however, this can lead to downstream issues because sometimes
    # consumers read graph.nodes() which ends up being [A, B] and sometimes it's [B, A]
    # but they base some of their logic on the order of the nodes, so the order ends up being important
    # so we sort the nodes in the edge in a stable way, so that we always get the same order
    if not graph.is_directed():

        def _sort_source_target(edge):
            source, target, edge_data = edge
            if source > target:
                temp = source
                source = target
                target = temp
            return source, target, edge_data

        edges = [_sort_source_target(edge) for edge in edges]

    def _get_edge_key(source: Any, target: Any) -> str:
        return f"{source} -> {target}"

    edges = sorted(edges, key=lambda x: _get_edge_key(x[0], x[1]))

    fixed_graph.add_edges_from(edges)
    return fixed_graph


def stable_largest_connected_component(graph: nx.Graph) -> nx.Graph:
    """Return the largest connected component of the graph, with nodes and edges sorted in a stable way."""
    graph = graph.copy()
    graph = largest_connected_component(graph)
    graph = normalize_node_names(graph)
    return _stabilize_graph(graph)


def _compute_leiden_communities(
    graph: nx.Graph | nx.DiGraph,
    max_cluster_size: int,
    use_lcc: bool,
    seed=0xDEADBEEF,
) -> dict[int, dict[str, int]]:
    """Return Leiden root communities."""
    if use_lcc:
        graph = stable_largest_connected_component(graph)

    community_mapping = hierarchical_leiden(
        graph, max_cluster_size=max_cluster_size, random_seed=seed
    )
    results: dict[int, dict[str, int]] = {}
    for partition in community_mapping:
        results[partition.level] = results.get(partition.level, {})
        results[partition.level][partition.node] = partition.cluster

    return results


def run_leiden(
    graph: nx.Graph, args: dict[str, Any]
) -> dict[int, dict[str, list[str]]]:
    """Run method definition."""
    max_cluster_size = args.get("max_cluster_size", 10)
    use_lcc = args.get("use_lcc", True)
    if args.get("verbose", False):
        print(
            "Running leiden with max_cluster_size=%s, lcc=%s", max_cluster_size, use_lcc
        )

    node_id_to_community_map = _compute_leiden_communities(
        graph=graph,
        max_cluster_size=max_cluster_size,
        use_lcc=use_lcc,
        seed=args.get("seed", 0xDEADBEEF),
    )
    levels = args.get("levels")

    # If they don't pass in levels, use them all
    if levels is None:
        levels = sorted(node_id_to_community_map.keys())

    results_by_level: dict[int, dict[str, list[str]]] = {}
    for level in levels:
        result = {}
        results_by_level[level] = result
        for node_id, raw_community_id in node_id_to_community_map[level].items():
            community_id = str(raw_community_id)
            if community_id not in result:
                result[community_id] = []
            result[community_id].append(node_id)
    return results_by_level


Communities = list[tuple[int, str, list[str]]]


class GraphCommunityStrategyType(str, Enum):
    """GraphCommunityStrategyType class definition."""

    leiden = "leiden"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


# TODO: This should support str | nx.Graph as a graphml param
def apply_clustering(
    graphml: str, communities: Communities, level=0, seed=0xF001
) -> nx.Graph:
    """Apply clustering to a graphml string."""
    random = Random(seed)  # noqa S311
    graph = nx.parse_graphml(graphml)
    for community_level, community_id, nodes in communities:
        if level == community_level:
            for node in nodes:
                graph.nodes[node]["cluster"] = community_id
                graph.nodes[node]["level"] = level

    # add node degree
    for node_degree in graph.degree:
        graph.nodes[str(node_degree[0])]["degree"] = int(node_degree[1])

    # add node uuid and incremental record id (a human readable id used as reference in the final report)
    for index, node in enumerate(graph.nodes()):
        graph.nodes[node]["human_readable_id"] = index
        graph.nodes[node]["id"] = str(gen_uuid(random))

    # add ids to edges
    for index, edge in enumerate(graph.edges()):
        graph.edges[edge]["id"] = str(gen_uuid(random))
        graph.edges[edge]["human_readable_id"] = index
        graph.edges[edge]["level"] = level
    return graph


def run_layout(
    strategy: dict[str, Any], graphml_or_graph: str | nx.Graph
) -> Communities:
    """Run layout method definition."""
    graph = load_graph(graphml_or_graph)
    if len(graph.nodes) == 0:
        print("Graph has no nodes")
        return []

    clusters: dict[int, dict[str, list[str]]] = {}
    strategy_type = strategy.get("type", GraphCommunityStrategyType.leiden)
    match strategy_type:
        case GraphCommunityStrategyType.leiden:
            clusters = run_leiden(graph, strategy)
        case _:
            msg = f"Unknown clustering strategy {strategy_type}"
            raise ValueError(msg)

    results: Communities = []
    for level in clusters:
        for cluster_id, nodes in clusters[level].items():
            results.append((level, cluster_id, nodes))
    return results


def cluster_graph(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    strategy: dict[str, Any],
    column: str,
    to: str,
    level_to: str | None = None,
    **_kwargs,
) -> pd.DataFrame:
    output_df = input
    results = output_df[column].apply(lambda graph: run_layout(strategy, graph))

    community_map_to = "communities"
    output_df[community_map_to] = results

    level_to = level_to or f"{to}_level"
    output_df[level_to] = output_df.apply(
        lambda x: list({level for level, _, _ in x[community_map_to]}), axis=1
    )
    output_df[to] = [None] * len(output_df)

    # Go through each of the rows
    graph_level_pairs_column: list[list[tuple[int, str]]] = []
    for _, row in output_df.iterrows():
        levels = row[level_to]
        graph_level_pairs: list[tuple[int, str]] = []

        # For each of the levels, get the graph and add it to the list
        for level in levels:
            graph = "\n".join(
                nx.generate_graphml(
                    apply_clustering(
                        row[column],
                        row[community_map_to],
                        level,
                    )
                )
            )
            graph_level_pairs.append((level, graph))
        graph_level_pairs_column.append(graph_level_pairs)
    output_df[to] = graph_level_pairs_column

    # explode the list of (level, graph) pairs into separate rows
    output_df = output_df.explode(to, ignore_index=True)

    # split the (level, graph) pairs into separate columns
    # TODO: There is probably a better way to do this
    output_df[[level_to, to]] = pd.DataFrame(
        output_df[to].tolist(), index=output_df.index
    )

    # clean up the community map
    output_df.drop(columns=[community_map_to], inplace=True)

    return output_df



from .create_base_text_units import *
from .create_base_extracted_entities import *
from .create_summarized_entities import *
from .create_base_entity_graph import *
from .create_final_entities import *
from .create_final_nodes import *
from .create_final_relationships import *
from .join_text_units_to_entity_ids import *
from .join_text_units_to_relationship_ids import *
from .create_final_community_reports import *
from .create_final_text_units import *
from .create_final_communities import *
from .create_base_documents import *
from .create_final_documents import *


import df_ops
import pandas as pd
import os

from utils.save import parquet_table_load, parquet_table_save


def create_final_text_units(
    base_text_units_output: pd.DataFrame,
    join_text_units_to_entity_ids_output: pd.DataFrame,
    join_text_unit_id_to_relationship_ids_output: pd.DataFrame,
    query_output_dir: str = None,
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_final_text_units"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    dataset = df_ops.select(
        base_text_units_output,
        **{"columns": ["id", "chunk", "document_ids", "n_tokens"]},
    )

    pre_entity_join = df_ops.rename(
        dataset,
        **{
            "columns": {
                "chunk": "text",
            },
        },
    )

    pre_relationship_join = df_ops.join(
        pre_entity_join,
        join_text_units_to_entity_ids_output,
        **{
            "on": ["id", "id"],
            "strategy": "left outer",
        },
    )

    pre_covariate_join = df_ops.join(
        pre_relationship_join,
        join_text_unit_id_to_relationship_ids_output,
        **{
            "on": ["id", "id"],
            "strategy": "left outer",
        },
    )

    dataset = df_ops.aggregate_override(
        pre_covariate_join,
        **{
            "groupby": ["id"],  # from the join above
            "aggregations": [
                {
                    "column": "text",
                    "operation": "any",
                    "to": "text",
                },
                {
                    "column": "n_tokens",
                    "operation": "any",
                    "to": "n_tokens",
                },
                {
                    "column": "document_ids",
                    "operation": "any",
                    "to": "document_ids",
                },
                {
                    "column": "entity_ids",
                    "operation": "any",
                    "to": "entity_ids",
                },
                {
                    "column": "relationship_ids",
                    "operation": "any",
                    "to": "relationship_ids",
                },
            ],
        },
    )

    final_text_units_output = df_ops.select(
        dataset,
        # Final select to get output in the correct shape
        columns=[
            "id",
            "text",
            "n_tokens",
            "document_ids",
            "entity_ids",
            "relationship_ids",
        ],
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, final_text_units_output)

    return final_text_units_output



import pandas as pd
from typing import Any, cast
from enum import Enum
from dataclasses import dataclass
import networkx as nx
from utils.graph import load_graph


DEFAULT_CONCAT_SEPARATOR = ","


class NumericOperation(str, Enum):
    """Numeric Operation class definition."""

    Sum = "sum"
    Average = "average"
    Max = "max"
    Min = "min"
    Multiply = "multiply"
    Replace = "replace"
    Skip = "skip"


class StringOperation(str, Enum):
    """String Operation class definition."""

    Concat = "concat"
    Replace = "replace"
    Skip = "skip"


class BasicMergeOperation(str, Enum):
    """Basic Merge Operation class definition."""

    Replace = "replace"
    Skip = "skip"


DEFAULT_NODE_OPERATIONS = {
    "*": {
        "operation": BasicMergeOperation.Replace,
    }
}

DEFAULT_EDGE_OPERATIONS = {
    "*": {
        "operation": BasicMergeOperation.Replace,
    },
    "weight": "sum",
}


@dataclass
class DetailedAttributeMergeOperation:
    """Detailed attribute merge operation class definition."""

    operation: StringOperation | NumericOperation

    # concat
    separator: str | None = None
    delimiter: str | None = None
    distinct: bool = False


def _get_detailed_attribute_merge_operation(
    value: str | dict[str, Any],
) -> DetailedAttributeMergeOperation:
    """Normalize the AttributeMergeOperation into a DetailedAttributeMergeOperation."""
    if isinstance(value, str):
        return DetailedAttributeMergeOperation(operation=value)
    return DetailedAttributeMergeOperation(**value)


def merge_edges(
    target_graph: nx.Graph,
    subgraph: nx.Graph,
    edge_ops: dict[str, DetailedAttributeMergeOperation],
):
    """Merge edges from subgraph into target using the operations defined in edge_ops."""
    for source, target, edge_data in subgraph.edges(data=True):  # type: ignore
        if not target_graph.has_edge(source, target):
            target_graph.add_edge(source, target, **(edge_data or {}))
        else:
            merge_attributes(target_graph.edges[(source, target)], edge_data, edge_ops)


def merge_nodes(
    target: nx.Graph,
    subgraph: nx.Graph,
    node_ops: dict[str, DetailedAttributeMergeOperation],
):
    """Merge nodes from subgraph into target using the operations defined in node_ops."""
    for node in subgraph.nodes:
        if node not in target.nodes:
            target.add_node(node, **(subgraph.nodes[node] or {}))
        else:
            merge_attributes(target.nodes[node], subgraph.nodes[node], node_ops)


def merge_graphs(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    column: str,
    to: str,
    nodes: dict[str, Any] = DEFAULT_NODE_OPERATIONS,
    edges: dict[str, Any] = DEFAULT_EDGE_OPERATIONS,
    **_kwargs,
) -> pd.DataFrame:

    input_df = input
    output = pd.DataFrame()

    node_ops = {
        attrib: _get_detailed_attribute_merge_operation(value)
        for attrib, value in nodes.items()
    }
    edge_ops = {
        attrib: _get_detailed_attribute_merge_operation(value)
        for attrib, value in edges.items()
    }

    mega_graph = nx.Graph()
    for graphml in input_df[column]:
        graph = load_graph(cast(str | nx.Graph, graphml))
        merge_nodes(mega_graph, graph, node_ops)
        merge_edges(mega_graph, graph, edge_ops)

    output[to] = ["\n".join(nx.generate_graphml(mega_graph))]

    return output


def apply_merge_operation(
    target_item: dict[str, Any] | None,
    source_item: dict[str, Any] | None,
    attrib: str,
    op: DetailedAttributeMergeOperation,
):
    """Apply the merge operation to the attribute."""
    source_item = source_item or {}
    target_item = target_item or {}

    if (
        op.operation == BasicMergeOperation.Replace
        or op.operation == StringOperation.Replace
    ):
        target_item[attrib] = source_item.get(attrib, None) or ""
    elif (
        op.operation == BasicMergeOperation.Skip or op.operation == StringOperation.Skip
    ):
        target_item[attrib] = target_item.get(attrib, None) or ""
    elif op.operation == StringOperation.Concat:
        separator = op.separator or DEFAULT_CONCAT_SEPARATOR
        target_attrib = target_item.get(attrib, "") or ""
        source_attrib = source_item.get(attrib, "") or ""
        target_item[attrib] = f"{target_attrib}{separator}{source_attrib}"
        if op.distinct:
            # TODO: Slow
            target_item[attrib] = separator.join(
                sorted(set(target_item[attrib].split(separator)))
            )

    # We're assuming that the attribute is numeric
    elif op.operation == NumericOperation.Sum:
        target_item[attrib] = (target_item.get(attrib, 0) or 0) + (
            source_item.get(attrib, 0) or 0
        )
    elif op.operation == NumericOperation.Average:
        target_item[attrib] = (
            (target_item.get(attrib, 0) or 0) + (source_item.get(attrib, 0) or 0)
        ) / 2
    elif op.operation == NumericOperation.Max:
        target_item[attrib] = max(
            (target_item.get(attrib, 0) or 0), (source_item.get(attrib, 0) or 0)
        )
    elif op.operation == NumericOperation.Min:
        target_item[attrib] = min(
            (target_item.get(attrib, 0) or 0), (source_item.get(attrib, 0) or 0)
        )
    elif op.operation == NumericOperation.Multiply:
        target_item[attrib] = (target_item.get(attrib, 1) or 1) * (
            source_item.get(attrib, 1) or 1
        )
    else:
        msg = f"Invalid operation {op.operation}"
        raise ValueError(msg)


def merge_attributes(
    target_item: dict[str, Any] | None,
    source_item: dict[str, Any] | None,
    ops: dict[str, DetailedAttributeMergeOperation],
):
    """Merge attributes from source_item into target_item using the operations defined in ops."""
    source_item = source_item or {}
    target_item = target_item or {}
    for op_attrib, op in ops.items():
        if op_attrib == "*":
            for attrib in source_item:
                # If there is a specific handler for this attribute, use it
                # i.e. * provides a default, but you can override it
                if attrib not in ops:
                    apply_merge_operation(target_item, source_item, attrib, op)
        else:
            if op_attrib in source_item or op_attrib in target_item:
                apply_merge_operation(target_item, source_item, op_attrib, op)



import pandas as pd
import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save


def join_text_units_to_entity_ids(
    create_final_entities_output: pd.DataFrame,
    query_output_dir: str,
    save: bool = True,
    try_load: bool = True,
) -> pd.DataFrame:

    fn_name = "join_text_units_to_entity_ids"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    dataset = df_ops.select(
        create_final_entities_output,
        **{"columns": ["id", "text_unit_ids"]},
    )
    dataset = df_ops.unroll(dataset, column="text_unit_ids")
    dataset = df_ops.aggregate_override(
        dataset,
        **{
            "groupby": ["text_unit_ids"],
            "aggregations": [
                {
                    "column": "id",
                    "operation": "array_agg_distinct",
                    "to": "entity_ids",
                },
                {
                    "column": "text_unit_ids",
                    "operation": "any",
                    "to": "id",
                },
            ],
        },
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset



from utils.hash import gen_md5_hash
import pandas as pd


def genid(
    input: pd.DataFrame,
    to: str,
    method: str = "md5_hash",
    hash: list[str] = [],  # noqa A002
) -> pd.DataFrame:
    """
    Generate a unique id for each row in the tabular data.

    ## Usage
    ### json
    ```json
    {
        "verb": "genid",
        "args": {
            "to": "id_output_column_name", /* The name of the column to output the id to */
            "method": "md5_hash", /* The method to use to generate the id */
            "hash": ["list", "of", "column", "names"] /* only if using md5_hash */,
            "seed": 034324 /* The random seed to use with UUID */
        }
    }
    ```

    ### yaml
    ```yaml
    verb: genid
    args:
        to: id_output_column_name
        method: md5_hash
        hash:
            - list
            - of
            - column
            - names
        seed: 034324
    ```
    """
    data = input

    if method == "md5_hash":
        if len(hash) == 0:
            msg = 'Must specify the "hash" columns to use md5_hash method'
            raise ValueError(msg)

        data[to] = data.apply(lambda row: gen_md5_hash(row, hash), axis=1)
    elif method == "increment":
        data[to] = data.index + 1
    else:
        msg = f"Unknown method {method}"
        raise ValueError(msg)
    return data



from typing import Any, cast

import networkx as nx
import pandas as pd
from utils.graph import load_graph

default_copy = ["level"]

def unpack_graph(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    column: str,
    type: str,  # noqa A002
    copy: list[str] | None = ["level"],
    embeddings_column: str = "embeddings",
    **kwargs,
) -> pd.DataFrame:
    """
    Unpack nodes or edges from a graphml graph, into a list of nodes or edges.

    This verb will create columns for each attribute in a node or edge.

    ## Usage
    ```yaml
    verb: unpack_graph
    args:
        type: node # The type of data to unpack, one of: node, edge. node will create a node list, edge will create an edge list
        column: <column name> # The name of the column containing the graph, should be a graphml graph
    ```
    """

    input_df = input
    # num_total = len(input_df)
    result = []
    copy = [col for col in copy if col in input_df.columns]
    has_embeddings = embeddings_column in input_df.columns

    for _, row in input_df.iterrows():
        # merge the original row with the unpacked graph item
        cleaned_row = {col: row[col] for col in copy}
        embeddings = (
            cast(dict[str, list[float]], row[embeddings_column])
            if has_embeddings
            else {}
        )

        result.extend(
            [
                {**cleaned_row, **graph_id}
                for graph_id in _run_unpack(
                    cast(str | nx.Graph, row[column]),
                    type,
                    embeddings,
                    kwargs,
                )
            ]
        )

    output_df = pd.DataFrame(result)
    return output_df


def _run_unpack(
    graphml_or_graph: str | nx.Graph,
    unpack_type: str,
    embeddings: dict[str, list[float]],
    args: dict[str, Any],
) -> list[dict[str, Any]]:
    graph = load_graph(graphml_or_graph)
    if unpack_type == "nodes":
        return _unpack_nodes(graph, embeddings, args)
    if unpack_type == "edges":
        return _unpack_edges(graph, args)
    msg = f"Unknown type {unpack_type}"
    raise ValueError(msg)


def _unpack_nodes(
    graph: nx.Graph, embeddings: dict[str, list[float]], _args: dict[str, Any]
) -> list[dict[str, Any]]:
    return [
        {
            "label": label,
            **(node_data or {}),
            "graph_embedding": embeddings.get(label),
        }
        for label, node_data in graph.nodes(data=True)  # type: ignore
    ]


def _unpack_edges(graph: nx.Graph, _args: dict[str, Any]) -> list[dict[str, Any]]:
    return [
        {
            "source": source_id,
            "target": target_id,
            **(edge_data or {}),
        }
        for source_id, target_id, edge_data in graph.edges(data=True)  # type: ignore
    ]



# Copyright (c) 2024 Microsoft Corporation.
# Licensed under the MIT License
"""Common field name definitions for community reports."""

# POST-PREP NODE TABLE SCHEMA
NODE_ID = "human_readable_id"
NODE_NAME = "title"
NODE_DESCRIPTION = "description"
NODE_DEGREE = "degree"
NODE_DETAILS = "node_details"
NODE_COMMUNITY = "community"
NODE_LEVEL = "level"

# POST-PREP EDGE TABLE SCHEMA
EDGE_ID = "human_readable_id"
EDGE_SOURCE = "source"
EDGE_TARGET = "target"
EDGE_DESCRIPTION = "description"
EDGE_DEGREE = "rank"
EDGE_DETAILS = "edge_details"
EDGE_WEIGHT = "weight"

# POST-PREP CLAIM TABLE SCHEMA
CLAIM_ID = "human_readable_id"
CLAIM_SUBJECT = "subject_id"
CLAIM_TYPE = "type"
CLAIM_STATUS = "status"
CLAIM_DESCRIPTION = "description"
CLAIM_DETAILS = "claim_details"

# COMMUNITY HIERARCHY TABLE SCHEMA
SUB_COMMUNITY = "sub_communitty"
SUB_COMMUNITY_SIZE = "sub_community_size"
COMMUNITY_LEVEL = "level"

# COMMUNITY CONTEXT TABLE SCHEMA
ALL_CONTEXT = "all_context"
CONTEXT_STRING = "context_string"
CONTEXT_SIZE = "context_size"
CONTEXT_EXCEED_FLAG = "context_exceed_limit"

# COMMUNITY REPORT TABLE SCHEMA
REPORT_ID = "id"
COMMUNITY_ID = "id"
COMMUNITY_LEVEL = "level"
TITLE = "title"
SUMMARY = "summary"
FINDINGS = "findings"
RATING = "rank"
EXPLANATION = "rating_explanation"
FULL_CONTENT = "full_content"
FULL_CONTENT_JSON = "full_content_json"



import os

from typing import Dict
from llm.send import ChatLLM
from df_ops.entity_extract import entity_extract
from df_ops.merge_graphs import merge_graphs
import pandas as pd
from template.graph_extract import GRAPH_EXTRACTION_PROMPT
from utils.save import parquet_table_load, parquet_table_save


def create_base_extracted_entities(
    dataset: pd.DataFrame,
    query_output_dir: str,
    llm_send_to: ChatLLM,
    llm_args: Dict = {},
    save: bool = True,
    try_load: bool = True,
):
    fn_name = "create_base_extracted_entities"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)

    dataset = entity_extract(
        input=dataset,
        send_to=llm_send_to,
        column="chunk",
        id_column="chunk_id",
        graph_to="entity_graph",
        to="entities",
        entity_types=["disease", "symptom", "cause"],
        extraction_prompt=GRAPH_EXTRACTION_PROMPT,
        llm_args=llm_args,
    )

    dataset = merge_graphs(
        dataset,
        column="entity_graph",
        to="entity_graph",
        **{
            "nodes": {
                "source_id": {
                    "operation": "concat",
                    "delimiter": ", ",
                    "distinct": True,
                },
                "description": (
                    {
                        "operation": "concat",
                        "separator": "\n",
                        "distinct": False,
                    }
                ),
            },
            "edges": {
                "source_id": {
                    "operation": "concat",
                    "delimiter": ", ",
                    "distinct": True,
                },
                "description": (
                    {
                        "operation": "concat",
                        "separator": "\n",
                        "distinct": False,
                    }
                ),
                "weight": "sum",
            },
        },
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)
    return dataset



import pandas as pd


def text_split_df(
    input: pd.DataFrame, column: str, to: str, separator: str = ","
) -> pd.DataFrame:
    """Split a column into a list of strings."""
    output = input

    def _apply_split(row):
        if row[column] is None or isinstance(row[column], list):
            return row[column]
        if row[column] == "":
            return []
        if not isinstance(row[column], str):
            message = f"Expected {column} to be a string, but got {type(row[column])}"
            raise TypeError(message)
        return row[column].split(separator)

    output[to] = output.apply(_apply_split, axis=1)
    return output


def text_split(
    input: pd.DataFrame,
    column: str,
    to: str,
    separator: str = ",",
) -> pd.DataFrame:
    """
    Split a piece of text into a list of strings based on a delimiter. The verb outputs a new column containing a list of strings.

    ## Usage

    ```yaml
    verb: text_split
    args:
        column: text # The name of the column containing the text to split
        to: split_text # The name of the column to output the split text to
        separator: "," # The separator to split the text on, defaults to ","
    ```
    """
    output = text_split_df(input, column, to, separator)
    return output



from .orderby import *
from .aggregate import *
from .zip import *
from .chunk import *
from .select import *
from .unroll import *
from .rename import *
from .genid import *
from .unzip import *
from .copy import *
from .filter import *
from .summarize_descriptions import *
from .cluster_graph import *
from .unpack import *
from .dedupe import *
from .text_split import *
from .drop import *
from .merge import *
from .text_embed import *
from .layout_graph import *
from .convert import *
from .join import *
from .concat import *
from .fill import *
from .compute_edge_combined_degree import *
from .prepare_community_reports import *
from .create_community_reports import *
from .restore_community_hierarchy import *
from .window import *



import pandas as pd


def unzip(
    input: pd.DataFrame,
    column: str,
    to: list[str],
) -> pd.DataFrame:
    """Unpacks a column containing a tuple into multiple columns."""
    table = input

    table[to] = pd.DataFrame(table[column].tolist(), index=table.index)

    return table



import pandas as pd
import tiktoken
from . import schemas

_MISSING_DESCRIPTION = "No Description"


def prepare_community_reports_nodes(
    input: pd.DataFrame,
    to: str = schemas.NODE_DETAILS,
    id_column: str = schemas.NODE_ID,
    name_column: str = schemas.NODE_NAME,
    description_column: str = schemas.NODE_DESCRIPTION,
    degree_column: str = schemas.NODE_DEGREE,
    **_kwargs,
) -> pd.DataFrame:
    """Merge edge details into an object."""
    node_df = input
    node_df = node_df.fillna(value={description_column: _MISSING_DESCRIPTION})

    # merge values of four columns into a map column
    node_df[to] = node_df.apply(
        lambda x: {
            id_column: x[id_column],
            name_column: x[name_column],
            description_column: x[description_column],
            degree_column: x[degree_column],
        },
        axis=1,
    )
    return node_df


def prepare_community_reports_edges(
    input: pd.DataFrame,
    to: str = schemas.EDGE_DETAILS,
    id_column: str = schemas.EDGE_ID,
    source_column: str = schemas.EDGE_SOURCE,
    target_column: str = schemas.EDGE_TARGET,
    description_column: str = schemas.EDGE_DESCRIPTION,
    degree_column: str = schemas.EDGE_DEGREE,
    **_kwargs,
) -> pd.DataFrame:
    """Merge edge details into an object."""
    edge_df: pd.DataFrame = input.fillna(
        value={description_column: _MISSING_DESCRIPTION}
    )
    edge_df[to] = edge_df.apply(
        lambda x: {
            id_column: x[id_column],
            source_column: x[source_column],
            target_column: x[target_column],
            description_column: x[description_column],
            degree_column: x[degree_column],
        },
        axis=1,
    )
    return edge_df


def get_levels(df: pd.DataFrame, level_column: str = schemas.NODE_LEVEL) -> list[int]:
    """Get the levels of the communities."""
    result = sorted(df[level_column].fillna(-1).unique().tolist(), reverse=True)
    return [r for r in result if r != -1]


def prepare_community_reports(
    node_df: pd.DataFrame,
    edge_df: pd.DataFrame,
    claim_df: pd.DataFrame | None = None,
    # callbacks: VerbCallbacks,
    max_tokens: int = 16_000,
    **_kwargs,
) -> pd.DataFrame:
    """Generate entities for each row, and optionally a graph of those entities."""
    # Prepare Community Reports
    # node_df = cast(pd.DataFrame, get_required_input_table(input, "nodes").table)
    # edge_df = cast(pd.DataFrame, get_required_input_table(input, "edges").table)
    # claim_df = get_named_input_table(input, "claims")
    if claim_df is not None:
        claim_df = claim_df

    levels = get_levels(node_df, schemas.NODE_LEVEL)
    dfs = []

    for level in levels:
        communities_at_level_df = _prepare_reports_at_level(
            node_df, edge_df, claim_df, level, max_tokens
        )
        dfs.append(communities_at_level_df)

    # build initial local context for all communities
    return pd.concat(dfs)


def filter_nodes_to_level(node_df: pd.DataFrame, level: int) -> pd.DataFrame:
    """Filter nodes to level."""
    return node_df[node_df[schemas.NODE_LEVEL] == level]


def filter_edges_to_nodes(edge_df: pd.DataFrame, nodes: list[str]) -> pd.DataFrame:
    """Filter edges to nodes."""
    return edge_df[
        edge_df[schemas.EDGE_SOURCE].isin(nodes)
        & edge_df[schemas.EDGE_TARGET].isin(nodes)
    ]


def filter_claims_to_nodes(claims_df: pd.DataFrame, nodes: list[str]) -> pd.DataFrame:
    """Filter edges to nodes."""
    return claims_df[claims_df[schemas.CLAIM_SUBJECT].isin(nodes)]


def num_tokens(text: str, token_encoder: tiktoken.Encoding | None = None) -> int:
    """Return the number of tokens in the given text."""
    if token_encoder is None:
        token_encoder = tiktoken.get_encoding("cl100k_base")
    return len(token_encoder.encode(text))  # type: ignore


def sort_context(
    local_context: list[dict],
    sub_community_reports: list[dict] | None = None,
    max_tokens: int | None = None,
    node_id_column: str = schemas.NODE_ID,
    node_name_column: str = schemas.NODE_NAME,
    node_details_column: str = schemas.NODE_DETAILS,
    edge_id_column: str = schemas.EDGE_ID,
    edge_details_column: str = schemas.EDGE_DETAILS,
    edge_degree_column: str = schemas.EDGE_DEGREE,
    edge_source_column: str = schemas.EDGE_SOURCE,
    edge_target_column: str = schemas.EDGE_TARGET,
    claim_id_column: str = schemas.CLAIM_ID,
    claim_details_column: str = schemas.CLAIM_DETAILS,
    community_id_column: str = schemas.COMMUNITY_ID,
) -> str:
    """Sort context by degree in descending order.

    If max tokens is provided, we will return the context string that fits within the token limit.
    """

    def _get_context_string(
        entities: list[dict],
        edges: list[dict],
        claims: list[dict],
        sub_community_reports: list[dict] | None = None,
    ) -> str:
        """Concatenate structured data into a context string."""
        contexts = []
        if sub_community_reports:
            sub_community_reports = [
                report
                for report in sub_community_reports
                if community_id_column in report
                and report[community_id_column]
                and str(report[community_id_column]).strip() != ""
            ]
            report_df = pd.DataFrame(sub_community_reports).drop_duplicates()
            if not report_df.empty:
                if report_df[community_id_column].dtype == float:
                    report_df[community_id_column] = report_df[
                        community_id_column
                    ].astype(int)
                report_string = (
                    f"----Reports-----\n{report_df.to_csv(index=False, sep=',')}"
                )
                contexts.append(report_string)

        entities = [
            entity
            for entity in entities
            if node_id_column in entity
            and entity[node_id_column]
            and str(entity[node_id_column]).strip() != ""
        ]
        entity_df = pd.DataFrame(entities).drop_duplicates()
        if not entity_df.empty:
            if entity_df[node_id_column].dtype == float:
                entity_df[node_id_column] = entity_df[node_id_column].astype(int)
            entity_string = (
                f"-----Entities-----\n{entity_df.to_csv(index=False, sep=',')}"
            )
            contexts.append(entity_string)

        if claims and len(claims) > 0:
            claims = [
                claim
                for claim in claims
                if claim_id_column in claim
                and claim[claim_id_column]
                and str(claim[claim_id_column]).strip() != ""
            ]
            claim_df = pd.DataFrame(claims).drop_duplicates()
            if not claim_df.empty:
                if claim_df[claim_id_column].dtype == float:
                    claim_df[claim_id_column] = claim_df[claim_id_column].astype(int)
                claim_string = (
                    f"-----Claims-----\n{claim_df.to_csv(index=False, sep=',')}"
                )
                contexts.append(claim_string)

        edges = [
            edge
            for edge in edges
            if edge_id_column in edge
            and edge[edge_id_column]
            and str(edge[edge_id_column]).strip() != ""
        ]
        edge_df = pd.DataFrame(edges).drop_duplicates()
        if not edge_df.empty:
            if edge_df[edge_id_column].dtype == float:
                edge_df[edge_id_column] = edge_df[edge_id_column].astype(int)
            edge_string = (
                f"-----Relationships-----\n{edge_df.to_csv(index=False, sep=',')}"
            )
            contexts.append(edge_string)

        return "\n\n".join(contexts)

    # sort node details by degree in descending order
    edges = []
    node_details = {}
    claim_details = {}

    for record in local_context:
        node_name = record[node_name_column]
        record_edges = record.get(edge_details_column, [])
        record_edges = [e for e in record_edges if not pd.isna(e)]
        record_node_details = record[node_details_column]
        record_claims = record.get(claim_details_column, [])
        record_claims = [c for c in record_claims if not pd.isna(c)]

        edges.extend(record_edges)
        node_details[node_name] = record_node_details
        claim_details[node_name] = record_claims

    edges = [edge for edge in edges if isinstance(edge, dict)]
    edges = sorted(edges, key=lambda x: x[edge_degree_column], reverse=True)

    sorted_edges = []
    sorted_nodes = []
    sorted_claims = []
    context_string = ""
    for edge in edges:
        source_details = node_details.get(edge[edge_source_column], {})
        target_details = node_details.get(edge[edge_target_column], {})
        sorted_nodes.extend([source_details, target_details])
        sorted_edges.append(edge)
        source_claims = claim_details.get(edge[edge_source_column], [])
        target_claims = claim_details.get(edge[edge_target_column], [])
        sorted_claims.extend(source_claims if source_claims else [])
        sorted_claims.extend(target_claims if source_claims else [])
        if max_tokens:
            new_context_string = _get_context_string(
                sorted_nodes, sorted_edges, sorted_claims, sub_community_reports
            )
            if num_tokens(context_string) > max_tokens:
                break
            context_string = new_context_string

    if context_string == "":
        return _get_context_string(
            sorted_nodes, sorted_edges, sorted_claims, sub_community_reports
        )

    return context_string


def set_context_size(df: pd.DataFrame) -> None:
    """Measure the number of tokens in the context."""
    df[schemas.CONTEXT_SIZE] = df[schemas.CONTEXT_STRING].apply(lambda x: num_tokens(x))


def set_context_exceeds_flag(df: pd.DataFrame, max_tokens: int) -> None:
    """Set a flag to indicate if the context exceeds the limit."""
    df[schemas.CONTEXT_EXCEED_FLAG] = df[schemas.CONTEXT_SIZE].apply(
        lambda x: x > max_tokens
    )


def _prepare_reports_at_level(
    node_df: pd.DataFrame,
    edge_df: pd.DataFrame,
    claim_df: pd.DataFrame | None,
    level: int,
    max_tokens: int = 16_000,
    community_id_column: str = schemas.COMMUNITY_ID,
    node_id_column: str = schemas.NODE_ID,
    node_name_column: str = schemas.NODE_NAME,
    node_details_column: str = schemas.NODE_DETAILS,
    node_level_column: str = schemas.NODE_LEVEL,
    node_degree_column: str = schemas.NODE_DEGREE,
    node_community_column: str = schemas.NODE_COMMUNITY,
    edge_id_column: str = schemas.EDGE_ID,
    edge_source_column: str = schemas.EDGE_SOURCE,
    edge_target_column: str = schemas.EDGE_TARGET,
    edge_degree_column: str = schemas.EDGE_DEGREE,
    edge_details_column: str = schemas.EDGE_DETAILS,
    claim_id_column: str = schemas.CLAIM_ID,
    claim_subject_column: str = schemas.CLAIM_SUBJECT,
    claim_details_column: str = schemas.CLAIM_DETAILS,
):
    def get_edge_details(node_df: pd.DataFrame, edge_df: pd.DataFrame, name_col: str):
        return node_df.merge(
            edge_df[[name_col, schemas.EDGE_DETAILS]].rename(
                columns={name_col: schemas.NODE_NAME}
            ),
            on=schemas.NODE_NAME,
            how="left",
        )

    level_node_df = filter_nodes_to_level(node_df, level)
    # log.info("Number of nodes at level=%s => %s", level, len(level_node_df))
    nodes = level_node_df[node_name_column].tolist()

    # Filter edges & claims to those containing the target nodes
    level_edge_df = filter_edges_to_nodes(edge_df, nodes)
    level_claim_df = (
        filter_claims_to_nodes(claim_df, nodes) if claim_df is not None else None
    )

    # concat all edge details per node
    merged_node_df = pd.concat(
        [
            get_edge_details(level_node_df, level_edge_df, edge_source_column),
            get_edge_details(level_node_df, level_edge_df, edge_target_column),
        ],
        axis=0,
    )
    merged_node_df = (
        merged_node_df.groupby(
            [
                node_name_column,
                node_community_column,
                node_degree_column,
                node_level_column,
            ]
        )
        .agg({node_details_column: "first", edge_details_column: list})
        .reset_index()
    )

    # concat claim details per node
    if level_claim_df is not None:
        merged_node_df = merged_node_df.merge(
            level_claim_df[[claim_subject_column, claim_details_column]].rename(
                columns={claim_subject_column: node_name_column}
            ),
            on=node_name_column,
            how="left",
        )
    merged_node_df = (
        merged_node_df.groupby(
            [
                node_name_column,
                node_community_column,
                node_level_column,
                node_degree_column,
            ]
        )
        .agg(
            {
                node_details_column: "first",
                edge_details_column: "first",
                **({claim_details_column: list} if level_claim_df is not None else {}),
            }
        )
        .reset_index()
    )

    # concat all node details, including name, degree, node_details, edge_details, and claim_details
    merged_node_df[schemas.ALL_CONTEXT] = merged_node_df.apply(
        lambda x: {
            node_name_column: x[node_name_column],
            node_degree_column: x[node_degree_column],
            node_details_column: x[node_details_column],
            edge_details_column: x[edge_details_column],
            claim_details_column: (
                x[claim_details_column] if level_claim_df is not None else []
            ),
        },
        axis=1,
    )

    # group all node details by community
    community_df = (
        merged_node_df.groupby(node_community_column)
        .agg({schemas.ALL_CONTEXT: list})
        .reset_index()
    )
    community_df[schemas.CONTEXT_STRING] = community_df[schemas.ALL_CONTEXT].apply(
        lambda x: sort_context(
            x,
            node_id_column=node_id_column,
            node_name_column=node_name_column,
            node_details_column=node_details_column,
            edge_id_column=edge_id_column,
            edge_details_column=edge_details_column,
            edge_degree_column=edge_degree_column,
            edge_source_column=edge_source_column,
            edge_target_column=edge_target_column,
            claim_id_column=claim_id_column,
            claim_details_column=claim_details_column,
            community_id_column=community_id_column,
        )
    )
    set_context_size(community_df)
    set_context_exceeds_flag(community_df, max_tokens)

    community_df[schemas.COMMUNITY_LEVEL] = level
    return community_df



import pandas as pd


def zip_verb(
    input: pd.DataFrame,
    to: str,
    columns: list[str],
    type: str | None = None,  # noqa A002
) -> pd.DataFrame:
    """Zip columns together"""
    table = input
    if type is None:
        table[to] = list(zip(*[table[col] for col in columns], strict=True))

    # This one is a little weird
    elif type == "dict":
        if len(columns) != 2:
            msg = f"Expected exactly two columns for a dict, got {columns}"
            raise ValueError(msg)
        key_col, value_col = columns

        results = []
        for _, row in table.iterrows():
            keys = row[key_col]
            values = row[value_col]
            output = {}
            if len(keys) != len(values):
                msg = f"Expected same number of keys and values, got {len(keys)} keys and {len(values)} values"
                raise ValueError(msg)
            for idx, key in enumerate(keys):
                output[key] = values[idx]
            results.append(output)

        table[to] = results
    return table.reset_index(drop=True)



import pandas as pd
from enum import Enum
from typing import Any, Iterable, cast, Callable
import nltk
from dataclasses import dataclass
import df_ops.defaults as defs
import tiktoken


initialized_nltk = False


class ChunkStrategyType(str, Enum):
    """ChunkStrategy class definition."""

    tokens = "tokens"
    sentence = "sentence"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


@dataclass
class TextChunk:
    """Text chunk class definition."""

    text_chunk: str
    source_doc_indices: list[int]
    n_tokens: int | None = None


EncodedText = list[int]
DecodeFn = Callable[[EncodedText], str]
EncodeFn = Callable[[str], EncodedText]
ChunkStrategy = Callable[[list[str], dict[str, Any]], Iterable[TextChunk]]
ChunkInput = str | list[str] | list[tuple[str, str]]


@dataclass(frozen=True)
class Tokenizer:
    """Tokenizer data class."""

    chunk_overlap: int
    """Overlap in tokens between chunks"""
    tokens_per_chunk: int
    """Maximum number of tokens per chunk"""
    decode: DecodeFn
    """ Function to decode a list of token ids to a string"""
    encode: EncodeFn
    """ Function to encode a string to a list of token ids"""


def split_text_on_tokens(
    texts: list[str],
    enc: Tokenizer,
) -> list[TextChunk]:
    """Split incoming text and return chunks."""
    result = []
    mapped_ids = []

    for source_doc_idx, text in enumerate(texts):
        encoded = enc.encode(text)
        mapped_ids.append((source_doc_idx, encoded))

    input_ids: list[tuple[int, int]] = [
        (source_doc_idx, id) for source_doc_idx, ids in mapped_ids for id in ids
    ]

    start_idx = 0
    cur_idx = min(start_idx + enc.tokens_per_chunk, len(input_ids))
    chunk_ids = input_ids[start_idx:cur_idx]
    while start_idx < len(input_ids):
        chunk_text = enc.decode([id for _, id in chunk_ids])
        doc_indices = list({doc_idx for doc_idx, _ in chunk_ids})
        result.append(
            TextChunk(
                text_chunk=chunk_text,
                source_doc_indices=doc_indices,
                n_tokens=len(chunk_ids),
            )
        )
        start_idx += enc.tokens_per_chunk - enc.chunk_overlap
        cur_idx = min(start_idx + enc.tokens_per_chunk, len(input_ids))
        chunk_ids = input_ids[start_idx:cur_idx]

    return result


def run_tokens(
    input: list[str],
    args: dict[str, Any],
) -> Iterable[TextChunk]:
    """Chunks text into multiple parts. A pipeline verb."""
    tokens_per_chunk = args.get("chunk_size", defs.CHUNK_SIZE)
    chunk_overlap = args.get("chunk_overlap", defs.CHUNK_OVERLAP)
    encoding_name = args.get("encoding_name", defs.ENCODING_MODEL)
    enc = tiktoken.get_encoding(encoding_name)

    def encode(text: str) -> list[int]:
        if not isinstance(text, str):
            text = f"{text}"
        return enc.encode(text)

    def decode(tokens: list[int]) -> str:
        return enc.decode(tokens)

    return split_text_on_tokens(
        input,
        Tokenizer(
            chunk_overlap=chunk_overlap,
            tokens_per_chunk=tokens_per_chunk,
            encode=encode,
            decode=decode,
        ),
    )


def bootstrap():
    """Bootstrap definition."""
    global initialized_nltk
    if not initialized_nltk:
        import nltk
        from nltk.corpus import wordnet as wn

        nltk.download("punkt")
        nltk.download("averaged_perceptron_tagger")
        nltk.download("maxent_ne_chunker")
        nltk.download("words")
        nltk.download("wordnet")
        wn.ensure_loaded()
        initialized_nltk = True


def run_sentence(input: list[str], _args: dict[str, Any]) -> Iterable[TextChunk]:
    """Chunks text into multiple parts. A pipeline verb."""
    for doc_idx, text in enumerate(input):
        sentences = nltk.sent_tokenize(text)
        for sentence in sentences:
            yield TextChunk(
                text_chunk=sentence,
                source_doc_indices=[doc_idx],
            )


def load_strategy(strategy: ChunkStrategyType) -> ChunkStrategy:
    """Load strategy method definition."""
    match strategy:
        case ChunkStrategyType.tokens:
            return run_tokens
        case ChunkStrategyType.sentence:
            # NLTK
            bootstrap()
            return run_sentence
        case _:
            msg = f"Unknown strategy: {strategy}"
            raise ValueError(msg)


def run_strategy(
    strategy: ChunkStrategy,
    input: ChunkInput,
    strategy_args: dict[str, Any],
) -> list[str | tuple[list[str] | None, str, int]]:
    """Run strategy method definition."""
    if isinstance(input, str):
        return [item.text_chunk for item in strategy([input], {**strategy_args})]

    # We can work with both just a list of text content
    # or a list of tuples of (document_id, text content)
    # text_to_chunk = '''
    texts = []
    for item in input:
        if isinstance(item, str):
            texts.append(item)
        else:
            texts.append(item[1])

    strategy_results = strategy(texts, {**strategy_args})

    results = []
    for strategy_result in strategy_results:
        doc_indices = strategy_result.source_doc_indices
        if isinstance(input[doc_indices[0]], str):
            results.append(strategy_result.text_chunk)
        else:
            doc_ids = [input[doc_idx][0] for doc_idx in doc_indices]
            results.append(
                (
                    doc_ids,
                    strategy_result.text_chunk,
                    strategy_result.n_tokens,
                )
            )
    return results


def chunk(
    input: pd.DataFrame,
    column: str,
    to: str,
    strategy: dict[str, Any] | None = None,
) -> pd.DataFrame:
    """
    Chunk a piece of text into smaller pieces.
    """
    if strategy is None:
        strategy = {}
    output = input
    strategy_name = strategy.get("type", ChunkStrategyType.tokens)
    strategy_config = {**strategy}
    strategy_exec = load_strategy(strategy_name)

    output[to] = output.apply(
        cast(
            Any,
            lambda x: run_strategy(strategy_exec, x[column], strategy_config),
        ),
        axis=1,
    )
    return output



import pandas as pd
from typing import Any

def concat(
    table: pd.DataFrame, others: list[pd.DataFrame], **_kwargs: Any
) -> pd.DataFrame:
    """Concat verb implementation."""
    return pd.concat([table] + others, ignore_index=True)



import pandas as pd
from llm.send import ChatLLM
from template.graph_extract import GRAPH_EXTRACTION_PROMPT
from . import defaults as defs
from dataclasses import dataclass
from generators.graph_extactor import GraphExtractor
from typing import Any, Dict
from text.text_splitter import create_text_splitter
import networkx as nx


@dataclass
class Document:
    """Document class definition."""

    text: str
    id: str


DEFAULT_ENTITY_TYPES = ["disease", "symptom"]
EntityTypes = list[str]
ExtractedEntity = dict[str, Any]


@dataclass
class EntityExtractionResult:
    """Entity extraction result class definition."""

    entities: list[ExtractedEntity]
    graphml_graph: str | None


def run_extract_entities(
    send_to: ChatLLM,
    docs: list[Document],
    entity_types: EntityTypes,
    # default params
    prechunked=True,
    chunk_size=1200,
    chunk_overlap=100,
    extraction_prompt=GRAPH_EXTRACTION_PROMPT,
    encoding_model="cl100k_base",
    encoding_name="cl100k_base",
    max_gleanings=1,
    tuple_delimiter=None,
    record_delimiter=None,
    completion_delimiter=None,
    llm_args: Dict | None = {},
) -> EntityExtractionResult:
    """Run the entity extraction chain."""

    text_splitter = create_text_splitter(
        prechunked, chunk_size, chunk_overlap, encoding_name
    )

    extractor = GraphExtractor(
        send_to=send_to,
        prompt=extraction_prompt,
        encoding_model=encoding_model,
        max_gleanings=max_gleanings,
        llm_args=llm_args,
    )

    text_list = [doc.text.strip() for doc in docs]

    # If it's not pre-chunked, then re-chunk the input
    if not prechunked:
        text_list = text_splitter.split_text("\n".join(text_list))

    results = extractor(
        texts=list(text_list),
        # send_to=send_to,
        prompt_variables={
            "entity_types": entity_types,
            "tuple_delimiter": tuple_delimiter,
            "record_delimiter": record_delimiter,
            "completion_delimiter": completion_delimiter,
        },
        # extraction_prompt=extraction_prompt,
        # max_gleaning=max_gleanings,
    )

    graph = results.output
    # Map the "source_id" back to the "id" field
    for _, node in graph.nodes(data=True):  # type: ignore
        if node is not None:
            node["source_id"] = ",".join(
                docs[int(id)].id for id in node["source_id"].split(",")
            )

    for _, _, edge in graph.edges(data=True):  # type: ignore
        if edge is not None:
            edge["source_id"] = ",".join(
                docs[int(id)].id for id in edge["source_id"].split(",")
            )

    entities = [
        ({"name": item[0], **(item[1] or {})})
        for item in graph.nodes(data=True)
        if item is not None
    ]

    graph_data = "".join(nx.generate_graphml(graph))
    return EntityExtractionResult(entities, graph_data)


def entity_extract(
    input: pd.DataFrame,
    send_to: ChatLLM,
    column: str,
    id_column: str,
    to: str,
    graph_to: str | None = None,
    entity_types=defs.DEFAULT_ENTITY_TYPES,
    extraction_prompt=GRAPH_EXTRACTION_PROMPT,
    llm_args: Dict | None = {},
) -> pd.DataFrame:

    if entity_types is None:
        entity_types = defs.DEFAULT_ENTITY_TYPES
    output = input
    num_started = 0

    def run_strategy(row):  # -> how the text is extracted?
        nonlocal num_started
        text = row[column]
        id = row[id_column]
        result = run_extract_entities(  # calling run_gi in graph intelligence.
            docs=[Document(text=text, id=id)],
            entity_types=entity_types,
            # callbacks,
            # cache,
            send_to=send_to,
            extraction_prompt=extraction_prompt,
            llm_args=llm_args,
        )
        num_started += 1
        return [result.entities, result.graphml_graph]

    results = [run_strategy(row) for _, row in output.iterrows()]

    to_result = []
    graph_to_result = []
    for result in results:
        if result:
            to_result.append(result[0])
            graph_to_result.append(result[1])
        else:
            to_result.append(None)
            graph_to_result.append(None)

    output[to] = to_result
    if graph_to is not None:
        output[graph_to] = graph_to_result

    return output.reset_index(drop=True)



from typing import Any, Dict, NamedTuple
from llm.send import ChatLLM

import pandas as pd
import networkx as nx

from dataclasses import dataclass
from utils.graph import load_graph
from generators.summarize_extractor import SummarizeExtractor


class DescriptionSummarizeRow(NamedTuple):
    """DescriptionSummarizeRow class definition."""

    graph: Any


@dataclass
class SummarizedDescriptionResult:
    """Entity summarization result class definition."""

    items: str | tuple[str, str]
    description: str


def run_summarize_descriptions(
    send_to: ChatLLM,
    items: str | tuple[str, str],
    descriptions: list[str],
    # reporter: VerbCallbacks,
    entity_name_key: str = "entity_name",
    input_descriptions_key: str = "description_list",
    summarize_prompt: str | None = None,
    max_tokens=4000,
    max_summary_length=500,
    llm_args: Dict = {},
) -> SummarizedDescriptionResult:
    """Run the entity extraction chain."""
    # Extraction Arguments

    extractor = SummarizeExtractor(
        send_to=send_to,
        summarization_prompt=summarize_prompt,
        entity_name_key=entity_name_key,
        input_descriptions_key=input_descriptions_key,
        # on_error=lambda e, stack, details: (
        #     reporter.error("Entity Extraction Error", e, stack, details)
        #     if reporter
        #     else None
        # ),
        max_summary_length=max_summary_length,
        max_input_tokens=max_tokens,
        llm_args=llm_args,
    )

    result = extractor(items=items, descriptions=descriptions)
    return SummarizedDescriptionResult(
        items=result.items, description=result.description
    )


def summarize_descriptions(
    input: pd.DataFrame,
    # cache: PipelineCache,
    # callbacks: VerbCallbacks,
    column: str,
    to: str,
    send_to: ChatLLM,
    max_summary_length: int = 500,
    # strategy: dict[str, Any] | None = None,
    llm_args: Dict | None = None,
) -> pd.DataFrame:
    output = input
    # strategy = strategy or {}
    # strategy_config = {**strategy}

    def get_resolved_entities(
        row,
    ):
        graph: nx.Graph = load_graph(getattr(row, column))
        # ticker_length = len(graph.nodes) + len(graph.edges)
        # ticker = progress_ticker(callbacks.progress, ticker_length)

        futures = [
            do_summarize_descriptions(
                node,
                sorted(set(graph.nodes[node].get("description", "").split("\n"))),
                # ticker,
            )
            for node in graph.nodes()
        ]
        futures += [
            do_summarize_descriptions(
                edge,
                sorted(set(graph.edges[edge].get("description", "").split("\n"))),
                # ticker,
            )
            for edge in graph.edges()
        ]

        results = futures

        for result in results:
            graph_item = result.items
            if isinstance(graph_item, str) and graph_item in graph.nodes():
                graph.nodes[graph_item]["description"] = result.description
            elif isinstance(graph_item, tuple) and graph_item in graph.edges():
                graph.edges[graph_item]["description"] = result.description

        return DescriptionSummarizeRow(
            graph="\n".join(nx.generate_graphml(graph)),
        )

    def do_summarize_descriptions(
        graph_item: str | tuple[str, str],
        descriptions: list[str],
        # ticker: ProgressTicker,
        # semaphore: asyncio.Semaphore,
    ):
        results = run_summarize_descriptions(
            max_summary_length=max_summary_length,
            send_to=send_to,
            items=graph_item,
            descriptions=descriptions,
            llm_args=llm_args,
        )
        return results

    # Graph is always on row 0, so here a derive from rows does not work
    # This iteration will only happen once, but avoids hardcoding a iloc[0]
    # Since parallelization is at graph level (nodes and edges), we can't use
    # the parallelization of the derive_from_rows
    # semaphore = asyncio.Semaphore(kwargs.get("num_threads", 4))

    results = [get_resolved_entities(row) for row in output.itertuples()]

    to_result = []

    for result in results:
        if result:
            to_result.append(result.graph)
        else:
            to_result.append(None)
    output[to] = to_result
    return output


StrategyConfig = dict[str, Any]



# Copyright (c) 2024 Microsoft Corporation.
# Licensed under the MIT License
from __future__ import annotations
from enum import Enum

"""Common default configuration values."""


# Copyright (c) 2024 Microsoft Corporation.
# Licensed under the MIT License

"""A module containing 'PipelineCacheConfig', 'PipelineFileCacheConfig' and 'PipelineMemoryCacheConfig' models."""


class AsyncType(str, Enum):
    """Enum for the type of async to use."""

    AsyncIO = "asyncio"
    Threaded = "threaded"


class CacheType(str, Enum):
    """The cache configuration type for the pipeline."""

    file = "file"
    """The file cache configuration type."""
    memory = "memory"
    """The memory cache configuration type."""
    none = "none"
    """The none cache configuration type."""
    blob = "blob"
    """The blob cache configuration type."""

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class InputFileType(str, Enum):
    """The input file type for the pipeline."""

    csv = "csv"
    """The CSV input type."""
    text = "text"
    """The text input type."""

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class InputType(str, Enum):
    """The input type for the pipeline."""

    file = "file"
    """The file storage type."""
    blob = "blob"
    """The blob storage type."""

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class StorageType(str, Enum):
    """The storage type for the pipeline."""

    file = "file"
    """The file storage type."""
    memory = "memory"
    """The memory storage type."""
    blob = "blob"
    """The blob storage type."""

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class ReportingType(str, Enum):
    """The reporting configuration type for the pipeline."""

    file = "file"
    """The file reporting configuration type."""
    console = "console"
    """The console reporting configuration type."""
    blob = "blob"
    """The blob reporting configuration type."""

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class TextEmbeddingTarget(str, Enum):
    """The target to use for text embeddings."""

    all = "all"
    required = "required"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


class LLMType(str, Enum):
    """LLMType enum class definition."""

    # Embeddings
    OpenAIEmbedding = "openai_embedding"
    AzureOpenAIEmbedding = "azure_openai_embedding"

    # Raw Completion
    OpenAI = "openai"
    AzureOpenAI = "azure_openai"

    # Chat Completion
    OpenAIChat = "openai_chat"
    AzureOpenAIChat = "azure_openai_chat"

    # Debug
    StaticResponse = "static_response"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


ASYNC_MODE = AsyncType.Threaded
ENCODING_MODEL = "cl100k_base"
#
# LLM Parameters
#
LLM_TYPE = LLMType.OpenAIChat
LLM_MODEL = "gpt-4-turbo-preview"
LLM_MAX_TOKENS = 4000
LLM_TEMPERATURE = 0
LLM_TOP_P = 1
LLM_N = 1
LLM_REQUEST_TIMEOUT = 180.0
LLM_TOKENS_PER_MINUTE = 0
LLM_REQUESTS_PER_MINUTE = 0
LLM_MAX_RETRIES = 10
LLM_MAX_RETRY_WAIT = 10.0
LLM_SLEEP_ON_RATE_LIMIT_RECOMMENDATION = True
LLM_CONCURRENT_REQUESTS = 25

#
# Text Embedding Parameters
#
EMBEDDING_TYPE = LLMType.OpenAIEmbedding
EMBEDDING_MODEL = "text-embedding-3-small"
EMBEDDING_BATCH_SIZE = 16
EMBEDDING_BATCH_MAX_TOKENS = 8191
EMBEDDING_TARGET = TextEmbeddingTarget.required

CACHE_TYPE = CacheType.file
CACHE_BASE_DIR = "cache"
CHUNK_SIZE = 1200
CHUNK_OVERLAP = 100
CHUNK_GROUP_BY_COLUMNS = ["id"]
CLAIM_DESCRIPTION = (
    "Any claims or facts that could be relevant to information discovery."
)
CLAIM_MAX_GLEANINGS = 1
CLAIM_EXTRACTION_ENABLED = False
MAX_CLUSTER_SIZE = 10
COMMUNITY_REPORT_MAX_LENGTH = 2000
COMMUNITY_REPORT_MAX_INPUT_LENGTH = 8000
ENTITY_EXTRACTION_ENTITY_TYPES = ["organization", "person", "geo", "event"]
ENTITY_EXTRACTION_MAX_GLEANINGS = 1
INPUT_FILE_TYPE = InputFileType.text
INPUT_TYPE = InputType.file
INPUT_BASE_DIR = "input"
INPUT_FILE_ENCODING = "utf-8"
INPUT_TEXT_COLUMN = "text"
INPUT_CSV_PATTERN = ".*\\.csv$"
INPUT_TEXT_PATTERN = ".*\\.txt$"
PARALLELIZATION_STAGGER = 0.3
PARALLELIZATION_NUM_THREADS = 50
NODE2VEC_ENABLED = False
NODE2VEC_NUM_WALKS = 10
NODE2VEC_WALK_LENGTH = 40
NODE2VEC_WINDOW_SIZE = 2
NODE2VEC_ITERATIONS = 3
NODE2VEC_RANDOM_SEED = 597832
REPORTING_TYPE = ReportingType.file
REPORTING_BASE_DIR = "output/${timestamp}/reports"
SNAPSHOTS_GRAPHML = False
SNAPSHOTS_RAW_ENTITIES = False
SNAPSHOTS_TOP_LEVEL_NODES = False
STORAGE_BASE_DIR = "output/${timestamp}/artifacts"
STORAGE_TYPE = StorageType.file
SUMMARIZE_DESCRIPTIONS_MAX_LENGTH = 500
UMAP_ENABLED = False

# Local Search
LOCAL_SEARCH_TEXT_UNIT_PROP = 0.5
LOCAL_SEARCH_COMMUNITY_PROP = 0.1
LOCAL_SEARCH_CONVERSATION_HISTORY_MAX_TURNS = 5
LOCAL_SEARCH_TOP_K_MAPPED_ENTITIES = 10
LOCAL_SEARCH_TOP_K_RELATIONSHIPS = 10
LOCAL_SEARCH_MAX_TOKENS = 12_000
LOCAL_SEARCH_LLM_TEMPERATURE = 0
LOCAL_SEARCH_LLM_TOP_P = 1
LOCAL_SEARCH_LLM_N = 1
LOCAL_SEARCH_LLM_MAX_TOKENS = 2000

# Global Search
GLOBAL_SEARCH_LLM_TEMPERATURE = 0
GLOBAL_SEARCH_LLM_TOP_P = 1
GLOBAL_SEARCH_LLM_N = 1
GLOBAL_SEARCH_MAX_TOKENS = 12_000
GLOBAL_SEARCH_DATA_MAX_TOKENS = 12_000
GLOBAL_SEARCH_MAP_MAX_TOKENS = 1000
GLOBAL_SEARCH_REDUCE_MAX_TOKENS = 2_000
GLOBAL_SEARCH_CONCURRENCY = 32

# Graph Extract
DEFAULT_TUPLE_DELIMITER = "<|>"
DEFAULT_RECORD_DELIMITER = "##"
DEFAULT_COMPLETION_DELIMITER = "<|COMPLETE|>"
DEFAULT_ENTITY_TYPES = ["disease", "symptoms", "cause"]



import pandas as pd
from typing import cast


def _degree_colname(column: str) -> str:
    return f"{column}_degree"


def _get_node_degree_table(
    nodes: pd.DataFrame, node_name_column: str, node_degree_column: str
) -> pd.DataFrame:
    return cast(pd.DataFrame, nodes[[node_name_column, node_degree_column]])


def compute_edge_combined_degree(
    input: pd.DataFrame,
    nodes: pd.DataFrame,
    to: str = "rank",
    node_name_column: str = "title",
    node_degree_column: str = "degree",
    edge_source_column: str = "source",
    edge_target_column: str = "target",
    **_kwargs,
) -> pd.DataFrame:
    """
    Compute the combined degree for each edge in a graph.

    Inputs Tables:
    - input: The edge table
    - nodes: The nodes table.

    Args:
    - to: The name of the column to output the combined degree to. Default="rank"
    """
    edge_df: pd.DataFrame = input
    if to in edge_df.columns:
        return edge_df
    node_degree_df = _get_node_degree_table(nodes, node_name_column, node_degree_column)

    def join_to_degree(df: pd.DataFrame, column: str) -> pd.DataFrame:
        degree_column = _degree_colname(column)
        result = df.merge(
            node_degree_df.rename(
                columns={node_name_column: column, node_degree_column: degree_column}
            ),
            on=column,
            how="left",
        )
        result[degree_column] = result[degree_column].fillna(0)
        return result

    edge_df = join_to_degree(edge_df, edge_source_column)
    edge_df = join_to_degree(edge_df, edge_target_column)
    edge_df[to] = (
        edge_df[_degree_colname(edge_source_column)]
        + edge_df[_degree_colname(edge_target_column)]
    )

    return edge_df



import pandas as pd
from typing import Any, cast
from typing_extensions import TypeAlias
from enum import Enum

Suffixes: TypeAlias = tuple[str | None, str | None]


class JoinStrategy(str, Enum):
    """Table join strategies."""

    Inner = "inner"
    LeftOuter = "left outer"
    RightOuter = "right outer"
    FullOuter = "full outer"
    AntiJoin = "anti join"
    SemiJoin = "semi join"
    Cross = "cross"


__strategy_mapping: dict[JoinStrategy, Any] = {
    JoinStrategy.Inner: "inner",
    JoinStrategy.LeftOuter: "left",
    JoinStrategy.RightOuter: "right",
    JoinStrategy.FullOuter: "outer",
    JoinStrategy.Cross: "cross",
    JoinStrategy.AntiJoin: "outer",
    JoinStrategy.SemiJoin: "outer",
}


def __clean_result(
    strategy: JoinStrategy, result: pd.DataFrame, source: pd.DataFrame
) -> pd.DataFrame:
    if strategy == JoinStrategy.AntiJoin:
        return cast(
            pd.DataFrame, result[result["_merge"] == "left_only"][source.columns]
        )
    if strategy == JoinStrategy.SemiJoin:
        return cast(pd.DataFrame, result[result["_merge"] == "both"][source.columns])

    result = cast(
        pd.DataFrame,
        pd.concat(
            [
                result[result["_merge"] == "both"],
                result[result["_merge"] == "left_only"],
                result[result["_merge"] == "right_only"],
            ]
        ),
    )
    return result.drop("_merge", axis=1)


def join(
    table: pd.DataFrame,
    other: pd.DataFrame,
    on: list[str] | None = None,
    strategy: str = "inner",
    **_kwargs: Any,
) -> pd.DataFrame:
    """Join verb implementation."""
    join_strategy = JoinStrategy(strategy)
    if on is not None and len(on) > 1:
        left_column = on[0]
        right_column = on[1]
        output = table.merge(
            other,
            left_on=left_column,
            right_on=right_column,
            how=__strategy_mapping[join_strategy],
            suffixes=cast(Suffixes, ["_1", "_2"]),
            indicator=True,
        )
    else:
        output = table.merge(
            other,
            on=on,
            how=__strategy_mapping[join_strategy],
            suffixes=cast(Suffixes, ["_1", "_2"]),
            indicator=True,
        )

    return __clean_result(join_strategy, output, table)



from typing import Dict
import pandas as pd

import df_ops
from llm.send import EmbLLM
import os
from utils.save import parquet_table_load, parquet_table_save

text_emb_strategy = {
    "batch_size": 16,
    "batch_max_tokens": 8191,
}


def create_final_entities(
    dataset: pd.DataFrame,
    query_output_dir: str,
    text_emb_llm_send_to: EmbLLM,
    text_emb_llm_args: Dict = {},
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_final_entities"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    dataset = df_ops.unpack_graph(
        dataset,
        column="clustered_graph",
        type="nodes",
    )

    dataset = df_ops.rename(
        dataset,
        columns={"label": "title"},
    )

    dataset = df_ops.select(
        dataset,
        columns=[
            "id",
            "title",
            "type",
            "description",
            "human_readable_id",
            "graph_embedding",
            "source_id",
        ],
    )

    dataset = df_ops.dedupe(dataset, columns=["id"])

    dataset = df_ops.rename(
        dataset,
        columns={"title": "name"},
    )

    dataset = df_ops.filter_verb(
        dataset,
        column="name",
        operator="is not empty",
        strategy="value",
        value=None,
    )

    dataset = df_ops.text_split(
        dataset,
        **{"separator": ",", "column": "source_id", "to": "text_unit_ids"},
    )

    dataset = df_ops.drop(dataset, columns=["source_id"])

    dataset = df_ops.merge(
        dataset,
        **{
            "strategy": "concat",
            "columns": ["name", "description"],
            "to": "name_description",
            "delimiter": ":",
            "preserveSource": True,
        },
    )

    dataset = df_ops.text_embed(
        input=dataset,
        send_to=text_emb_llm_send_to,
        llm_args=text_emb_llm_args,
        column="name_description",
        to="description_embedding",
        strategy=text_emb_strategy,
    )

    dataset = df_ops.drop(
        dataset,
        columns=["name_description"],
    )

    dataset = df_ops.filter_verb(
        chunk=dataset,
        column="description_embedding",
        operator="is not empty",
        strategy="value",
        value=None,
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)
    return dataset



import pandas as pd
import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save


def create_final_documents(
    base_documents_output: pd.DataFrame,
    query_output_dir: str | None = None,
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_final_documents"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    final_documents_output = df_ops.rename(
        base_documents_output, **{"columns": {"text_units": "text_unit_ids"}}
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, final_documents_output)
    return final_documents_output



import df_ops
import pandas as pd
import os
from utils.save import parquet_table_load, parquet_table_save
from typing import Dict

DEFAULT_CLUSTERING_STRATEGY = {"type": "leiden", "max_cluster_size": 10}


def create_base_entity_graph(
    dataset: pd.DataFrame,
    query_output_dir: str,
    clustering_strategy: Dict = DEFAULT_CLUSTERING_STRATEGY,
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_base_entity_graph"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    dataset = df_ops.cluster_graph(
        dataset,
        strategy=clustering_strategy,
        column="entity_graph",
        to="clustered_graph",
        level_to="level",
    )

    dataset = df_ops.select(
        dataset,
        columns=["level", "clustered_graph"],
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset



import pandas as pd 
from typing import Any

def fill(
    table: pd.DataFrame, to: str, value: str | float | bool, **_kwargs: Any
) -> pd.DataFrame:
    """Fill verb implementation."""
    table[to] = value
    return table



import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save
import pandas as pd


def create_final_communities(
    base_entity_graph_output: pd.DataFrame,
    query_output_dir: str | None = None,
    save=True,
    try_load: bool = True,
):
    fn_name = "create_final_communities"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    graph_nodes_output = df_ops.unpack_graph(
        base_entity_graph_output,
        **{
            "column": "clustered_graph",
            "type": "nodes",
        },
    )

    graph_edges_output = df_ops.unpack_graph(
        base_entity_graph_output,
        **{
            "column": "clustered_graph",
            "type": "edges",
        },
    )

    source_clusters_output = df_ops.join(
        table=graph_nodes_output,
        other=graph_edges_output,
        on=["label", "source"],
    )

    target_clusters_output = df_ops.join(
        table=graph_nodes_output,
        other=graph_edges_output,
        on=["label", "target"],
    )

    concatenated_clusters = df_ops.concat(
        table=source_clusters_output,
        others=[target_clusters_output],
    )

    combined_cluster = df_ops.filter_verb(
        concatenated_clusters,
        # level_1 is the left side of the join
        # level_2 is the right side of the join
        column="level_1",
        operator="equals",
        strategy="column",
        value="level_2",
    )

    cluster_relationships = df_ops.aggregate_override(
        combined_cluster,
        **{
            "groupby": [
                "cluster",
                "level_1",  # level_1 is the left side of the join
            ],
            "aggregations": [
                {
                    "column": "id_2",  # this is the id of the edge from the join steps above
                    "to": "relationship_ids",
                    "operation": "array_agg_distinct",
                },
                {
                    "column": "source_id_1",
                    "to": "text_unit_ids",
                    "operation": "array_agg_distinct",
                },
            ],
        },
    )

    all_clusters = df_ops.aggregate_override(
        graph_nodes_output,
        **{
            "groupby": ["cluster", "level"],
            "aggregations": [{"column": "cluster", "to": "id", "operation": "any"}],
        },
    )

    joined_dataset = df_ops.join(
        all_clusters, other=cluster_relationships, on=["id", "cluster"]
    )

    joined_dataset = df_ops.filter_verb(
        joined_dataset,
        strategy="column",
        column="level",
        operator="equals",
        value="level_1",
    )

    joined_dataset = df_ops.fill(
        joined_dataset,
        **{
            "to": "__temp",
            "value": "Community ",
        },
    )

    joined_dataset = df_ops.merge(
        joined_dataset,
        **{
            "columns": [
                "__temp",
                "id",
            ],
            "to": "title",
            "strategy": "concat",
            "preserveSource": True,
        },
    )

    joined_dataset = df_ops.copy(
        joined_dataset,
        **{
            "column": "id",
            "to": "raw_community",
        },
    )

    joined_dataset = df_ops.select(
        joined_dataset,
        **{
            "columns": [
                "id",
                "title",
                "level",
                "raw_community",
                "relationship_ids",
                "text_unit_ids",
            ],
        },
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, joined_dataset)

    return joined_dataset



from template.community_report import COMMUNITY_REPORT_PROMPT
from llm.send import ChatLLM, ModelArgs
import pandas as pd
import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save


def create_final_community_reports(
    final_nodes_output: pd.DataFrame,
    final_relationship_output: pd.DataFrame,
    query_output_dir: str,
    community_report_send_to: ChatLLM,
    community_report_llm_args: ModelArgs,
    save: bool = True,
    try_load: bool = True,
):
    fn_name = "create_final_community_reports"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)
    nodes = df_ops.prepare_community_reports_nodes(
        final_nodes_output,
    )
    edges = df_ops.prepare_community_reports_edges(
        final_relationship_output,
    )

    community_hierarchy = df_ops.restore_community_hierarchy(nodes)
    local_contexts = df_ops.prepare_community_reports(
        node_df=nodes,
        edge_df=edges,
        claim_df=None,
    )
    dataset = df_ops.create_community_reports(
        community_report_send_to=community_report_send_to,
        local_contexts=local_contexts,
        nodes=nodes,
        community_hierarchy=community_hierarchy,
        strategy={
            "extraction_prompt": COMMUNITY_REPORT_PROMPT,
            "max_report_length": 2000,
            "max_input_length": 8000,
        },
        community_report_llm_args=community_report_llm_args,
    )
    dataset = df_ops.window(
        dataset, **{"to": "id", "operation": "uuid", "column": "community"}
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)
    return dataset



from index.graphrag import GraphRAGIndexer
from index.vectorstore import VectorStoreIndexer
from llm.send import LLMSendToConfig
import argparse
import logging
import os
from llm.create import get_send_fn, get_text_emb_send_fn

# fmt: off
def get_args_parser():
    parser = argparse.ArgumentParser("Index documents", add_help=False)
    parser.add_argument("--output_dir", default="./index_results", type=str, help="Output directory for generated index files.",)
    parser.add_argument("--query", default=None, type=str, help="Query used for retrieving documents.")
    parser.add_argument("--doc_dir", default="./documents", type=str, help="Directory for saving retrieved documents.")
    parser.add_argument("--doc_top_k", default=10, type=int, help="Directory for saving retrieved documents.")
    parser.add_argument("--force", action="store_true", help="Run the index without local results (.parquets)." )
    parser.add_argument("--source", default="openai", type=str, help="Source of the LLM: [openai, huggingface]",)
    parser.add_argument("--text_emb_source", default="openai", type=str, help="Source of the text embedding LLM: [openai, huggingface]",)
    parser.add_argument("--model_name", default="gpt-3.5-turbo", help="Model to load for generation.")
    parser.add_argument("--text_emb_model_name", default="text-embedding-3-small", help="Model to load for text embedding.")
    parser.add_argument("--store_type", default="graphrag", help="Model to load for text embedding. Options: [graphrag,  ]")
    return parser
# fmt: on


def main(args):
    DEFAULT_LLM_ARGS = {
        "temperature": 0.0,
        "top_p": 1.0,
    }  # This setup for reproducibility.

    logging.basicConfig(
        # filename=f"{args.name}.log",
        level=logging.INFO,
    )

    send_to = get_send_fn(args.source, args.model_name)
    text_emb_send_to = get_text_emb_send_fn(
        args.text_emb_source, args.text_emb_model_name
    )

    if args.store_type == "graphrag":
        indexer = GraphRAGIndexer(
            doc_top_k=args.doc_top_k,
            final_entities_text_emb_llm_config=LLMSendToConfig(
                llm_send_to=text_emb_send_to,
                llm_model_args={},
            ),
            graph_extractor_llm_config=LLMSendToConfig(
                llm_send_to=send_to,
                llm_model_args=DEFAULT_LLM_ARGS,
            ),
            summarize_extractor_llm_config=LLMSendToConfig(
                llm_send_to=send_to,
                llm_model_args=DEFAULT_LLM_ARGS,
            ),
            community_report_llm_config=LLMSendToConfig(
                llm_send_to=send_to,
                llm_model_args=DEFAULT_LLM_ARGS,
            ),
            output_dir=os.path.join(args.output_dir, "graphrag"),
            doc_saving_dir=args.doc_dir,
        )

        indexer.generate(
            query=args.query,
            save=True,
            try_load=not args.force,
        )
    elif args.store_type == "vectorstore":
        indexer = VectorStoreIndexer(
            emb_llm_config=LLMSendToConfig(
                llm_send_to=text_emb_send_to,
                llm_model_args={},
            ),
            output_dir=os.path.join(args.output_dir, "vectorstore"),
            doc_saving_dir=args.doc_dir,
            doc_top_k=args.doc_top_k,
        )
        query_output_dir = indexer.generate(
            query=args.query,
            save=True,
        )

        indexer.save_emb_llm_info(
            query_output_dir=query_output_dir,
            source=args.text_emb_source,
            model_name=args.text_emb_model_name,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Indexing script", parents=[get_args_parser()])
    args = parser.parse_args()
    main(args)



from typing import Dict
import pandas as pd
from llm.send import ChatLLM
import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save


def create_summarized_entities(
    dataset: pd.DataFrame,
    query_output_dir: str,
    send_to: ChatLLM,
    llm_args: Dict = {},
    save: bool = True,
    try_load: bool = True,
):
    fn_name = "create_summarized_entities"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)

    dataset = df_ops.summarize_descriptions(
        input=dataset,
        send_to=send_to,
        column="entity_graph",
        to="entity_graph",
        max_summary_length=500,
        llm_args=llm_args,
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset



import json
import tiktoken
import pandas as pd


from . import schemas
from typing import Any, Callable
from llm.send import ChatLLM, ModelArgs
from typing_extensions import TypedDict
from generators.community_reports_extractor import CommunityReportsExtractor
from .select import select
from .join import join

COMMUNITY_REPORT_MAX_INPUT_LENGTH = 8000


class Finding(TypedDict):
    """Finding class definition."""

    summary: str
    explanation: str


def _parse_rank(report: dict) -> float:
    rank = report.get("rating", -1)
    try:
        return float(rank)
    except ValueError:
        print("Error parsing rank: %s defaulting to -1", rank)
        return -1


class CommunityReport(TypedDict):
    """Community report class definition."""

    community: str | int
    title: str
    summary: str
    full_content: str
    full_content_json: str
    rank: float
    level: int
    rank_explanation: str
    findings: list[Finding]


def get_levels(df: pd.DataFrame, level_column: str = schemas.NODE_LEVEL) -> list[int]:
    """Get the levels of the communities."""
    result = sorted(df[level_column].fillna(-1).unique().tolist(), reverse=True)
    return [r for r in result if r != -1]


def _run_extractor(
    send_to: ChatLLM,
    community: str | int,
    input: str,
    level: int,
    args: dict[str, Any],
    llm_args: ModelArgs,
    # reporter: VerbCallbacks,
) -> CommunityReport | None:
    # RateLimiter

    extractor = CommunityReportsExtractor(
        send_to,
        extraction_prompt=args.get("extraction_prompt", None),
        max_report_length=args.get("max_report_length", None),
        # on_error=lambda e, stack, _data: reporter.error(
        #     "Community Report Extraction Error", e, stack
        # ),
        llm_args=llm_args,
    )

    # try:
    results = extractor({"input_text": input})
    report = results.structured_output
    if report is None or len(report.keys()) == 0:
        print("No report found for community: %s", community)
        return None

    return CommunityReport(
        community=community,
        full_content=results.output,
        level=level,
        rank=_parse_rank(report),
        title=report.get("title", f"Community Report: {community}"),
        rank_explanation=report.get("rating_explanation", ""),
        summary=report.get("summary", ""),
        findings=report.get("findings", []),
        full_content_json=json.dumps(report, indent=4),
    )
    # except Exception as e:
    #     print("Error processing community: %s", community)
    #     return None


def where_column_equals(df: pd.DataFrame, column: str, value: Any) -> pd.DataFrame:
    """Return a filtered DataFrame where a column equals a value."""
    return df[df[column] == value]


def _at_level(level: int, df: pd.DataFrame) -> pd.DataFrame:
    """Return records at the given level."""
    return where_column_equals(df, schemas.COMMUNITY_LEVEL, level)


def _within_context(df: pd.DataFrame) -> pd.DataFrame:
    """Return records where the context is within the limit."""
    return where_column_equals(df, schemas.CONTEXT_EXCEED_FLAG, 0)


def _exceeding_context(df: pd.DataFrame) -> pd.DataFrame:
    """Return records where the context exceeds the limit."""
    return where_column_equals(df, schemas.CONTEXT_EXCEED_FLAG, 1)


def transform_series(series: pd.Series, fn: Callable[[Any], Any]) -> pd.Series:
    """Apply a transformation function to a series."""
    return series.apply(fn)


def num_tokens(text: str, token_encoder: tiktoken.Encoding | None = None) -> int:
    """Return the number of tokens in the given text."""
    if token_encoder is None:
        token_encoder = tiktoken.get_encoding("cl100k_base")
    return len(token_encoder.encode(text))  # type: ignore


def set_context_size(df: pd.DataFrame) -> None:
    """Measure the number of tokens in the context."""
    df[schemas.CONTEXT_SIZE] = df[schemas.CONTEXT_STRING].apply(lambda x: num_tokens(x))


def sort_context(
    local_context: list[dict],
    sub_community_reports: list[dict] | None = None,
    max_tokens: int | None = None,
    node_id_column: str = schemas.NODE_ID,
    node_name_column: str = schemas.NODE_NAME,
    node_details_column: str = schemas.NODE_DETAILS,
    edge_id_column: str = schemas.EDGE_ID,
    edge_details_column: str = schemas.EDGE_DETAILS,
    edge_degree_column: str = schemas.EDGE_DEGREE,
    edge_source_column: str = schemas.EDGE_SOURCE,
    edge_target_column: str = schemas.EDGE_TARGET,
    claim_id_column: str = schemas.CLAIM_ID,
    claim_details_column: str = schemas.CLAIM_DETAILS,
    community_id_column: str = schemas.COMMUNITY_ID,
) -> str:
    """Sort context by degree in descending order.

    If max tokens is provided, we will return the context string that fits within the token limit.
    """

    def _get_context_string(
        entities: list[dict],
        edges: list[dict],
        claims: list[dict],
        sub_community_reports: list[dict] | None = None,
    ) -> str:
        """Concatenate structured data into a context string."""
        contexts = []
        if sub_community_reports:
            sub_community_reports = [
                report
                for report in sub_community_reports
                if community_id_column in report
                and report[community_id_column]
                and str(report[community_id_column]).strip() != ""
            ]
            report_df = pd.DataFrame(sub_community_reports).drop_duplicates()
            if not report_df.empty:
                if report_df[community_id_column].dtype == float:
                    report_df[community_id_column] = report_df[
                        community_id_column
                    ].astype(int)
                report_string = (
                    f"----Reports-----\n{report_df.to_csv(index=False, sep=',')}"
                )
                contexts.append(report_string)

        entities = [
            entity
            for entity in entities
            if node_id_column in entity
            and entity[node_id_column]
            and str(entity[node_id_column]).strip() != ""
        ]
        entity_df = pd.DataFrame(entities).drop_duplicates()
        if not entity_df.empty:
            if entity_df[node_id_column].dtype == float:
                entity_df[node_id_column] = entity_df[node_id_column].astype(int)
            entity_string = (
                f"-----Entities-----\n{entity_df.to_csv(index=False, sep=',')}"
            )
            contexts.append(entity_string)

        if claims and len(claims) > 0:
            claims = [
                claim
                for claim in claims
                if claim_id_column in claim
                and claim[claim_id_column]
                and str(claim[claim_id_column]).strip() != ""
            ]
            claim_df = pd.DataFrame(claims).drop_duplicates()
            if not claim_df.empty:
                if claim_df[claim_id_column].dtype == float:
                    claim_df[claim_id_column] = claim_df[claim_id_column].astype(int)
                claim_string = (
                    f"-----Claims-----\n{claim_df.to_csv(index=False, sep=',')}"
                )
                contexts.append(claim_string)

        edges = [
            edge
            for edge in edges
            if edge_id_column in edge
            and edge[edge_id_column]
            and str(edge[edge_id_column]).strip() != ""
        ]
        edge_df = pd.DataFrame(edges).drop_duplicates()
        if not edge_df.empty:
            if edge_df[edge_id_column].dtype == float:
                edge_df[edge_id_column] = edge_df[edge_id_column].astype(int)
            edge_string = (
                f"-----Relationships-----\n{edge_df.to_csv(index=False, sep=',')}"
            )
            contexts.append(edge_string)

        return "\n\n".join(contexts)

    # sort node details by degree in descending order
    edges = []
    node_details = {}
    claim_details = {}

    for record in local_context:
        node_name = record[node_name_column]
        record_edges = record.get(edge_details_column, [])
        record_edges = [e for e in record_edges if not pd.isna(e)]
        record_node_details = record[node_details_column]
        record_claims = record.get(claim_details_column, [])
        record_claims = [c for c in record_claims if not pd.isna(c)]

        edges.extend(record_edges)
        node_details[node_name] = record_node_details
        claim_details[node_name] = record_claims

    edges = [edge for edge in edges if isinstance(edge, dict)]
    edges = sorted(edges, key=lambda x: x[edge_degree_column], reverse=True)

    sorted_edges = []
    sorted_nodes = []
    sorted_claims = []
    context_string = ""
    for edge in edges:
        source_details = node_details.get(edge[edge_source_column], {})
        target_details = node_details.get(edge[edge_target_column], {})
        sorted_nodes.extend([source_details, target_details])
        sorted_edges.append(edge)
        source_claims = claim_details.get(edge[edge_source_column], [])
        target_claims = claim_details.get(edge[edge_target_column], [])
        sorted_claims.extend(source_claims if source_claims else [])
        sorted_claims.extend(target_claims if source_claims else [])
        if max_tokens:
            new_context_string = _get_context_string(
                sorted_nodes, sorted_edges, sorted_claims, sub_community_reports
            )
            if num_tokens(context_string) > max_tokens:
                break
            context_string = new_context_string

    if context_string == "":
        return _get_context_string(
            sorted_nodes, sorted_edges, sorted_claims, sub_community_reports
        )

    return context_string


def _sort_and_trim_context(df: pd.DataFrame, max_tokens: int) -> pd.Series:
    """Sort and trim context to fit the limit."""
    series = df[schemas.ALL_CONTEXT]
    return transform_series(series, lambda x: sort_context(x, max_tokens=max_tokens))


def union(*frames: pd.DataFrame) -> pd.DataFrame:
    """Perform a union operation on the given set of dataframes."""
    return pd.concat(list(frames))


def antijoin(df: pd.DataFrame, exclude: pd.DataFrame, column: str) -> pd.DataFrame:
    """Return an anti-joined dataframe.

    Arguments:
    * df: The DataFrame to apply the exclusion to
    * exclude: The DataFrame containing rows to remove.
    * column: The join-on column.
    """
    result = df.merge(
        exclude[[column]],
        on=column,
        how="outer",
        indicator=True,
    )
    if "_merge" in result.columns:
        result = result[result["_merge"] == "left_only"].drop("_merge", axis=1)
    return result


def _antijoin_reports(df: pd.DataFrame, reports: pd.DataFrame) -> pd.DataFrame:
    """Return records in df that are not in reports."""
    return antijoin(df, reports, schemas.NODE_COMMUNITY)


def _get_subcontext_df(
    level: int, report_df: pd.DataFrame, local_context_df: pd.DataFrame
) -> pd.DataFrame:
    """Get sub-community context for each community."""
    sub_report_df = _drop_community_level(_at_level(level, report_df))
    sub_context_df = _at_level(level, local_context_df)
    sub_context_df = join(sub_context_df, sub_report_df, schemas.NODE_COMMUNITY)
    sub_context_df.rename(
        columns={schemas.NODE_COMMUNITY: schemas.SUB_COMMUNITY}, inplace=True
    )
    return sub_context_df


def drop_columns(df: pd.DataFrame, *column: str) -> pd.DataFrame:
    """Drop columns from a dataframe."""
    return df.drop(list(column), axis=1)


def _drop_community_level(df: pd.DataFrame) -> pd.DataFrame:
    """Drop the community level column from the dataframe."""
    return drop_columns(df, schemas.COMMUNITY_LEVEL)


def build_mixed_context(context: list[dict], max_tokens: int) -> str:
    """
    Build parent context by concatenating all sub-communities' contexts.

    If the context exceeds the limit, we use sub-community reports instead.
    """
    sorted_context = sorted(
        context, key=lambda x: x[schemas.CONTEXT_SIZE], reverse=True
    )

    # replace local context with sub-community reports, starting from the biggest sub-community
    substitute_reports = []
    final_local_contexts = []
    exceeded_limit = True
    context_string = ""

    for idx, sub_community_context in enumerate(sorted_context):
        if exceeded_limit:
            if sub_community_context[schemas.FULL_CONTENT]:
                substitute_reports.append(
                    {
                        schemas.COMMUNITY_ID: sub_community_context[
                            schemas.SUB_COMMUNITY
                        ],
                        schemas.FULL_CONTENT: sub_community_context[
                            schemas.FULL_CONTENT
                        ],
                    }
                )
            else:
                # this sub-community has no report, so we will use its local context
                final_local_contexts.extend(sub_community_context[schemas.ALL_CONTEXT])
                continue

            # add local context for the remaining sub-communities
            remaining_local_context = []
            for rid in range(idx + 1, len(sorted_context)):
                remaining_local_context.extend(sorted_context[rid][schemas.ALL_CONTEXT])
            new_context_string = sort_context(
                local_context=remaining_local_context + final_local_contexts,
                sub_community_reports=substitute_reports,
            )
            if num_tokens(new_context_string) <= max_tokens:
                exceeded_limit = False
                context_string = new_context_string
                break

    if exceeded_limit:
        # if all sub-community reports exceed the limit, we add reports until context is full
        substitute_reports = []
        for sub_community_context in sorted_context:
            substitute_reports.append(
                {
                    schemas.COMMUNITY_ID: sub_community_context[schemas.SUB_COMMUNITY],
                    schemas.FULL_CONTENT: sub_community_context[schemas.FULL_CONTENT],
                }
            )
            new_context_string = pd.DataFrame(substitute_reports).to_csv(
                index=False, sep=","
            )
            if num_tokens(new_context_string) > max_tokens:
                break

            context_string = new_context_string
    return context_string


def _build_mixed_context(df: pd.DataFrame, max_tokens: int) -> pd.Series:
    """Sort and trim context to fit the limit."""
    series = df[schemas.ALL_CONTEXT]
    return transform_series(
        series, lambda x: build_mixed_context(x, max_tokens=max_tokens)
    )


def _get_community_df(
    level: int,
    invalid_context_df: pd.DataFrame,
    sub_context_df: pd.DataFrame,
    community_hierarchy_df: pd.DataFrame,
    max_tokens: int,
) -> pd.DataFrame:
    """Get community context for each community."""
    # collect all sub communities' contexts for each community
    community_df = _drop_community_level(_at_level(level, community_hierarchy_df))
    invalid_community_ids = select(invalid_context_df, schemas.NODE_COMMUNITY)
    subcontext_selection = select(
        sub_context_df,
        schemas.SUB_COMMUNITY,
        schemas.FULL_CONTENT,
        schemas.ALL_CONTEXT,
        schemas.CONTEXT_SIZE,
    )

    invalid_communities = join(
        community_df, invalid_community_ids, schemas.NODE_COMMUNITY, "inner"
    )
    community_df = join(
        invalid_communities, subcontext_selection, schemas.SUB_COMMUNITY
    )
    community_df[schemas.ALL_CONTEXT] = community_df.apply(
        lambda x: {
            schemas.SUB_COMMUNITY: x[schemas.SUB_COMMUNITY],
            schemas.ALL_CONTEXT: x[schemas.ALL_CONTEXT],
            schemas.FULL_CONTENT: x[schemas.FULL_CONTENT],
            schemas.CONTEXT_SIZE: x[schemas.CONTEXT_SIZE],
        },
        axis=1,
    )
    community_df = (
        community_df.groupby(schemas.NODE_COMMUNITY)
        .agg({schemas.ALL_CONTEXT: list})
        .reset_index()
    )
    community_df[schemas.CONTEXT_STRING] = _build_mixed_context(
        community_df, max_tokens
    )
    community_df[schemas.COMMUNITY_LEVEL] = level
    return community_df


def prep_community_report_context(
    report_df: pd.DataFrame | None,
    community_hierarchy_df: pd.DataFrame,
    local_context_df: pd.DataFrame,
    level: int | str,
    max_tokens: int,
) -> pd.DataFrame:
    """
    Prep context for each community in a given level.

    For each community:
    - Check if local context fits within the limit, if yes use local context
    - If local context exceeds the limit, iteratively replace local context with sub-community reports, starting from the biggest sub-community
    """
    if report_df is None:
        report_df = pd.DataFrame()

    level = int(level)
    level_context_df = _at_level(level, local_context_df)
    valid_context_df = _within_context(level_context_df)
    invalid_context_df = _exceeding_context(level_context_df)

    # there is no report to substitute with, so we just trim the local context of the invalid context records
    # this case should only happen at the bottom level of the community hierarchy where there are no sub-communities
    if invalid_context_df.empty:
        return valid_context_df

    if report_df.empty:
        invalid_context_df[schemas.CONTEXT_STRING] = _sort_and_trim_context(
            invalid_context_df, max_tokens
        )
        set_context_size(invalid_context_df)
        invalid_context_df[schemas.CONTEXT_EXCEED_FLAG] = 0
        return union(valid_context_df, invalid_context_df)

    level_context_df = _antijoin_reports(level_context_df, report_df)

    # for each invalid context, we will try to substitute with sub-community reports
    # first get local context and report (if available) for each sub-community
    sub_context_df = _get_subcontext_df(level + 1, report_df, local_context_df)
    community_df = _get_community_df(
        level, invalid_context_df, sub_context_df, community_hierarchy_df, max_tokens
    )

    # handle any remaining invalid records that can't be subsituted with sub-community reports
    # this should be rare, but if it happens, we will just trim the local context to fit the limit
    remaining_df = _antijoin_reports(invalid_context_df, community_df)
    remaining_df[schemas.CONTEXT_STRING] = _sort_and_trim_context(
        remaining_df, max_tokens
    )

    result = union(valid_context_df, community_df, remaining_df)
    set_context_size(result)
    result[schemas.CONTEXT_EXCEED_FLAG] = 0
    return result


# def _generate_report(
#     send_to: Callable,
#     # cache: PipelineCache,
#     # callbacks: VerbCallbacks,
#     strategy: dict,
#     community_id: int | str,
#     community_level: int,
#     community_context: str,
# ) -> CommunityReport | None:
#     """Generate a report for a single community."""

#     return _run_extractor(
#         send_to,
#         community_id,
#         community_context,
#         community_level,
#         strategy,
#     )


def create_community_reports(
    community_report_send_to: ChatLLM,
    local_contexts: pd.DataFrame,
    nodes: pd.DataFrame,
    community_hierarchy: pd.DataFrame,
    # callbacks: VerbCallbacks,
    # cache: PipelineCache,
    strategy: dict,
    # async_mode: AsyncType = AsyncType.AsyncIO,
    # num_threads: int = 4,
    community_report_llm_args: ModelArgs = {},
    **_kwargs,
) -> pd.DataFrame:
    """Generate entities for each row, and optionally a graph of those entities."""
    # log.debug("create_community_reports strategy=%s", strategy)

    levels = get_levels(nodes)
    reports: list[CommunityReport | None] = []

    for level in levels:
        level_contexts = prep_community_report_context(
            pd.DataFrame(reports),
            local_context_df=local_contexts,
            community_hierarchy_df=community_hierarchy,
            level=level,
            max_tokens=strategy.get(
                "max_input_tokens", COMMUNITY_REPORT_MAX_INPUT_LENGTH
            ),
        )

        def run_generate(record):
            # result = _generate_report(
            #     create_community_send_to,
            #     _run_extractor,
            #     community_id=record[schemas.NODE_COMMUNITY],
            #     community_level=record[schemas.COMMUNITY_LEVEL],
            #     community_context=record[schemas.CONTEXT_STRING],
            #     # cache=cache,
            #     # callbacks=callbacks,
            #     strategy=strategy,
            # )
            result = _run_extractor(
                community_report_send_to,
                community=record[schemas.NODE_COMMUNITY],
                level=record[schemas.COMMUNITY_LEVEL],
                input=record[schemas.CONTEXT_STRING],
                args=strategy,
                llm_args=community_report_llm_args,
            )
            # tick()
            return result

        local_reports = []
        for _, record in level_contexts.iterrows():
            report = run_generate(record)
            local_reports.append(report)

        # local_reports = derive_from_rows(
        #     level_contexts,
        #     run_generate,
        #     callbacks=NoopVerbCallbacks(),
        #     num_threads=num_threads,
        #     scheduling_type=async_mode,
        # )
        reports.extend([lr for lr in local_reports if lr is not None])

    return pd.DataFrame(reports)



import pandas as pd
from typing import Any, cast, Callable
from enum import Enum
from functools import partial
from dataclasses import dataclass
from uuid import uuid4


class StringComparisonOperator(str, Enum):
    """String comparison operators."""

    Equals = "equals"
    NotEqual = "is not equal"
    Contains = "contains"
    StartsWith = "starts with"
    EndsWith = "ends with"
    IsEmpty = "is empty"
    IsNotEmpty = "is not empty"
    RegularExpression = "regex"


class NumericComparisonOperator(str, Enum):
    """Numeric comparison operators."""

    Equals = "="
    NotEqual = "!="
    LessThan = "<"
    LessThanOrEqual = "<="
    GreaterThan = ">"
    GreaterThanOrEqual = ">="
    IsEmpty = "is empty"
    IsNotEmpty = "is not empty"


class BooleanComparisonOperator(str, Enum):
    """Boolean comparison operators."""

    Equals = "equals"
    NotEqual = "is not equal"
    IsTrue = "is true"
    IsFalse = "is false"
    IsEmpty = "is empty"
    IsNotEmpty = "is not empty"


class ComparisonStrategy(str, Enum):
    """Filter compare type."""

    Value = "value"
    Column = "column"


class BooleanLogicalOperator(str, Enum):
    """Boolean logical operators."""

    OR = "or"
    AND = "and"
    NOR = "nor"
    NAND = "nand"
    XOR = "xor"
    XNOR = "xnor"


@dataclass
class InputColumnArgs:
    """Column argument for verbs operating on a single column."""

    column: str


@dataclass
class FilterArgs(InputColumnArgs):
    """Filter criteria for a column."""

    value: Any
    strategy: ComparisonStrategy
    operator: (
        NumericComparisonOperator | StringComparisonOperator | BooleanComparisonOperator
    )


def __correct_unknown_value(df: pd.DataFrame, columns: list[str], target: str) -> None:
    na_index = df[df[columns].isna().any(axis=1)].index
    df.loc[na_index, target] = None


def __equals(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column] == target


def __not_equals(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return ~df[column] == target


def __is_null(
    df: pd.DataFrame, column: str, **_kwargs: dict
) -> pd.DataFrame | pd.Series:
    return df[column].isna()


def __is_not_null(
    df: pd.DataFrame, column: str, **_kwargs: dict
) -> pd.DataFrame | pd.Series:
    return df[column].notna()


def __contains(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.DataFrame | pd.Series:
    return df[column].str.contains(str(target), regex=False)


def __startswith(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.DataFrame | pd.Series:
    return df[column].str.startswith(str(target))


def __endswith(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column].str.endswith(str(target))


def __regex(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column].str.contains(str(target), regex=True)


def __gt(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column] > target


def __gte(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column] >= target


def __lt(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column] < target


def __lte(
    df: pd.DataFrame,
    column: str,
    target: pd.Series | str | float | bool,
    **_kwargs: dict,
) -> pd.Series:
    return df[column] <= target


_operator_map: dict[
    StringComparisonOperator | NumericComparisonOperator | BooleanComparisonOperator,
    Callable,
] = {
    StringComparisonOperator.Contains: __contains,
    StringComparisonOperator.StartsWith: __startswith,
    StringComparisonOperator.EndsWith: __endswith,
    StringComparisonOperator.Equals: __equals,
    StringComparisonOperator.NotEqual: __not_equals,
    StringComparisonOperator.IsEmpty: __is_null,
    StringComparisonOperator.IsNotEmpty: __is_not_null,
    StringComparisonOperator.RegularExpression: __regex,
    NumericComparisonOperator.Equals: __equals,
    NumericComparisonOperator.IsEmpty: __is_null,
    NumericComparisonOperator.IsNotEmpty: __is_not_null,
    NumericComparisonOperator.GreaterThan: __gt,
    NumericComparisonOperator.GreaterThanOrEqual: __gte,
    NumericComparisonOperator.LessThan: __lt,
    NumericComparisonOperator.LessThanOrEqual: __lte,
    BooleanComparisonOperator.Equals: __equals,
    BooleanComparisonOperator.NotEqual: __not_equals,
    BooleanComparisonOperator.IsEmpty: __is_null,
    BooleanComparisonOperator.IsNotEmpty: __is_not_null,
    BooleanComparisonOperator.IsTrue: partial(__equals, target=True),
    BooleanComparisonOperator.IsFalse: partial(__equals, target=False),
}


boolean_function_map = {
    BooleanLogicalOperator.OR: lambda df, columns: (
        df[columns].any(axis="columns") if columns != "" else df.any(axis="columns")
    ),
    BooleanLogicalOperator.AND: lambda df, columns: (
        df[columns].all(axis="columns") if columns != "" else df.all(axis="columns")
    ),
    BooleanLogicalOperator.NOR: lambda df, columns: (
        ~df[columns].any(axis="columns") if columns != "" else ~df.any(axis="columns")
    ),
    BooleanLogicalOperator.NAND: lambda df, columns: (
        ~df[columns].all(axis="columns") if columns != "" else ~df.all(axis="columns")
    ),
    BooleanLogicalOperator.XNOR: lambda df, columns: (
        df[columns].sum(axis="columns").apply(lambda x: (x % 2) == 0 or x == 0)
        if columns != ""
        else df.sum(axis="columns").apply(lambda x: (x % 2) == 0 or x == 0)
    ),
    BooleanLogicalOperator.XOR: lambda df, columns: (
        df[columns].sum(axis="columns").apply(lambda x: (x % 2) != 0 and x != 0)
        if columns != ""
        else df.sum(axis="columns").apply(lambda x: (x % 2) != 0 and x != 0)
    ),
}


_empty_comparisons = {
    StringComparisonOperator.IsEmpty,
    StringComparisonOperator.IsNotEmpty,
    NumericComparisonOperator.IsEmpty,
    NumericComparisonOperator.IsNotEmpty,
    BooleanComparisonOperator.IsEmpty,
    BooleanComparisonOperator.IsNotEmpty,
}


def filter_fn(
    df: pd.DataFrame, args: FilterArgs
) -> pd.DataFrame | pd.Series:  # noqa A001 - use ds verb name
    """Filter a DataFrame based on the input criteria."""
    filters: list[str] = []
    filtered_df: pd.DataFrame = df.copy()

    filter_name = str(uuid4())
    filters.append(filter_name)
    if args.strategy == ComparisonStrategy.Column:
        filtered_df[filter_name] = _operator_map[args.operator](
            df=df, column=args.column, target=df[args.value]
        )
        if args.operator not in _empty_comparisons:
            __correct_unknown_value(filtered_df, [args.column, args.value], filter_name)
    else:
        filtered_df[filter_name] = _operator_map[args.operator](
            df=df, column=args.column, target=args.value
        )

    filtered_df["dwc_filter_result"] = boolean_function_map[BooleanLogicalOperator.OR](
        filtered_df[filters], ""
    )

    __correct_unknown_value(filtered_df, filters, "dwc_filter_result")

    return filtered_df["dwc_filter_result"]


def get_comparison_operator(
    operator: str,
) -> StringComparisonOperator | NumericComparisonOperator | BooleanComparisonOperator:
    """Get a comparison operator based on the input string."""
    try:
        return StringComparisonOperator(operator)
    except Exception:
        print("%s is not a string comparison operator", operator)
    try:
        return NumericComparisonOperator(operator)
    except Exception:
        print("%s is not a numeric comparison operator", operator)
    try:
        return BooleanComparisonOperator(operator)
    except Exception:
        print("%s is not a boolean comparison operator", operator)
    raise NotImplementedError(operator)


def filter_verb(
    chunk: pd.DataFrame,
    column: str,
    value: Any,
    strategy: ComparisonStrategy = ComparisonStrategy.Value,
    operator: StringComparisonOperator = StringComparisonOperator.Equals,
) -> pd.DataFrame:
    """Filter verb implementation."""
    input_table = cast(pd.DataFrame, chunk)

    filter_index = filter_fn(
        input_table,
        FilterArgs(
            column,
            value=value,
            strategy=ComparisonStrategy(strategy),
            operator=get_comparison_operator(operator),
        ),
    )
    sub_idx = filter_index == True  # noqa: E712
    idx = filter_index[sub_idx].index  # type: ignore
    result = input_table[chunk.index.isin(idx)].reset_index(drop=True)
    return result



from typing import List
import df_ops
import pandas as pd
import os
from utils.save import parquet_table_load, parquet_table_save


def create_base_text_units(
    dataset: pd.DataFrame,
    query_output_dir: str,
    chunk_column_name: str = "chunk",
    chunk_by_columns: List[str] = ["id"],
    n_tokens_column_name: str = "n_tokens",
    save: bool = True,
    try_load: bool = True,
):
    fn_name = "create_base_text_units"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)

    dataset = df_ops.orderby(dataset, [{"column": "id", "direction": "asc"}])
    dataset = df_ops.zip_verb(dataset, columns=["id", "text"], to="text_with_ids")
    dataset = df_ops.aggregate_override(
        dataset,
        groupby=([*chunk_by_columns] if len(chunk_by_columns) > 0 else None),
        aggregations=[
            {
                "column": "text_with_ids",
                "operation": "array_agg",
                "to": "texts",
            }
        ],
    )

    dataset = df_ops.chunk(
        dataset,
        column="texts",
        to="chunks",
        strategy={
            "type": "tokens",
            "chunk_size": 1200,
            "chunk_overlap": 100,
            "group_by_columns": ["id"],
        },
    )
    dataset = df_ops.select(
        dataset,
        columns=[*chunk_by_columns, "chunks"],
    )

    dataset = df_ops.unroll(
        dataset,
        column="chunks",
    )

    dataset = df_ops.rename(
        dataset,
        columns={
            "chunks": chunk_column_name,
        },
    )

    dataset = df_ops.genid(
        dataset,
        to="chunk_id",
        method="md5_hash",
        hash=[chunk_column_name],
    )

    dataset = df_ops.unzip(
        dataset,
        column=chunk_column_name,
        to=["document_ids", chunk_column_name, n_tokens_column_name],
    )

    dataset = df_ops.copy(
        dataset,
        to="id",
        column="chunk_id",
    )

    dataset = df_ops.filter_verb(
        dataset,
        column=chunk_column_name,
        value=None,
        strategy="value",
        operator="is not empty",
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset



from setuptools import setup, find_packages

setup(
    name="graphrag_for_all",
    version="0.1.3",
    description="Graphrag and vectorstore for all LLMs",
    url="https://github.com/ChihchengHsieh/rag-aug",
    author="Chihcheng Hsieh",
    author_email="chihcheng.hsieh.82@gmail.com",
    license="MIT",
    # packages=["graphrag_for_all"],
    packages=find_packages(),
    zip_safe=False,
)



import pandas as pd
from typing import Any

def dedupe(
    table: pd.DataFrame, columns: list[str] | None = None, **_kwargs: Any
) -> pd.DataFrame:
    """Dedupe verb implementation."""
    return table.drop_duplicates(columns)


from enum import Enum
from typing import Any, cast
import networkx as nx
import pandas as pd
from dataclasses import dataclass
from utils.graph import load_graph
import umap
import numpy as np


class LayoutGraphStrategyType(str, Enum):
    """LayoutGraphStrategyType class definition."""

    umap = "umap"
    zero = "zero"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


@dataclass
class NodePosition:
    """Node position class definition."""

    label: str
    cluster: str
    size: float

    x: float
    y: float
    z: float | None = None

    def to_pandas(self) -> tuple[str, float, float, str, float]:
        """To pandas method definition."""
        return self.label, self.x, self.y, self.cluster, self.size


GraphLayout = list[NodePosition]
NodeEmbeddings = dict[str, list[float]]


def layout_graph(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    strategy: dict[str, Any],
    embeddings_column: str,
    graph_column: str,
    to: str,
    graph_to: str | None = None,
    **_kwargs: dict,
) -> pd.DataFrame:

    output_df = input

    # num_items = len(output_df)
    strategy_type = strategy.get("type", LayoutGraphStrategyType.umap)
    strategy_args = {**strategy}

    has_embeddings = embeddings_column in output_df.columns

    layouts = output_df.apply(
        lambda row: _run_layout(
            strategy_type,
            row[graph_column],
            row[embeddings_column] if has_embeddings else {},
            strategy_args,
        ),
        axis=1,
    )
    output_df[to] = layouts.apply(lambda layout: [pos.to_pandas() for pos in layout])
    if graph_to is not None:
        output_df[graph_to] = output_df.apply(
            lambda row: _apply_layout_to_graph(
                row[graph_column], cast(GraphLayout, layouts[row.name])
            ),
            axis=1,
        )
    return output_df


def compute_umap_positions(
    embedding_vectors: np.ndarray,
    node_labels: list[str],
    node_categories: list[int] | None = None,
    node_sizes: list[int] | None = None,
    min_dist: float = 0.75,
    n_neighbors: int = 25,
    spread: int = 1,
    metric: str = "euclidean",
    n_components: int = 2,
    random_state: int = 86,
) -> list[NodePosition]:
    """Project embedding vectors down to 2D/3D using UMAP."""
    embedding_positions = umap.UMAP(
        min_dist=min_dist,
        n_neighbors=n_neighbors,
        spread=spread,
        n_components=n_components,
        metric=metric,
        random_state=random_state,
    ).fit_transform(embedding_vectors)

    embedding_position_data: list[NodePosition] = []
    for index, node_name in enumerate(node_labels):
        node_points = embedding_positions[index]  # type: ignore
        node_category = 1 if node_categories is None else node_categories[index]
        node_size = 1 if node_sizes is None else node_sizes[index]

        if len(node_points) == 2:
            embedding_position_data.append(
                NodePosition(
                    label=str(node_name),
                    x=float(node_points[0]),
                    y=float(node_points[1]),
                    cluster=str(int(node_category)),
                    size=int(node_size),
                )
            )
        else:
            embedding_position_data.append(
                NodePosition(
                    label=str(node_name),
                    x=float(node_points[0]),
                    y=float(node_points[1]),
                    z=float(node_points[2]),
                    cluster=str(int(node_category)),
                    size=int(node_size),
                )
            )
    return embedding_position_data


def run_umap(
    graph: nx.Graph,
    embeddings: NodeEmbeddings,
    args: dict[str, Any],
) -> GraphLayout:
    """Run method definition."""
    node_clusters = []
    node_sizes = []

    embeddings = _filter_raw_embeddings(embeddings)
    nodes = list(embeddings.keys())
    embedding_vectors = [embeddings[node_id] for node_id in nodes]

    for node_id in nodes:
        node = graph.nodes[node_id]
        cluster = node.get("cluster", node.get("community", -1))
        node_clusters.append(cluster)
        size = node.get("degree", node.get("size", 0))
        node_sizes.append(size)

    additional_args = {}
    if len(node_clusters) > 0:
        additional_args["node_categories"] = node_clusters
    if len(node_sizes) > 0:
        additional_args["node_sizes"] = node_sizes

    try:
        return compute_umap_positions(
            embedding_vectors=np.array(embedding_vectors),
            node_labels=nodes,
            **additional_args,
            min_dist=args.get("min_dist", 0.75),
            n_neighbors=args.get("n_neighbors", 5),
        )
    except Exception as e:
        print("Error running UMAP")
        # Umap may fail due to input sparseness or memory pressure.
        # For now, in these cases, we'll just return a layout with all nodes at (0, 0)
        result = []
        for i in range(len(nodes)):
            cluster = node_clusters[i] if len(node_clusters) > 0 else 1
            result.append(
                NodePosition(x=0, y=0, label=nodes[i], size=0, cluster=str(cluster))
            )
        return result


def _filter_raw_embeddings(embeddings: NodeEmbeddings) -> NodeEmbeddings:
    return {
        node_id: embedding
        for node_id, embedding in embeddings.items()
        if embedding is not None
    }


def get_zero_positions(
    node_labels: list[str],
    node_categories: list[int] | None = None,
    node_sizes: list[int] | None = None,
    three_d: bool | None = False,
) -> list[NodePosition]:
    """Project embedding vectors down to 2D/3D using UMAP."""
    embedding_position_data: list[NodePosition] = []
    for index, node_name in enumerate(node_labels):
        node_category = 1 if node_categories is None else node_categories[index]
        node_size = 1 if node_sizes is None else node_sizes[index]

        if not three_d:
            embedding_position_data.append(
                NodePosition(
                    label=str(node_name),
                    x=0,
                    y=0,
                    cluster=str(int(node_category)),
                    size=int(node_size),
                )
            )
        else:
            embedding_position_data.append(
                NodePosition(
                    label=str(node_name),
                    x=0,
                    y=0,
                    z=0,
                    cluster=str(int(node_category)),
                    size=int(node_size),
                )
            )
    return embedding_position_data


def run_zero(
    graph: nx.Graph,
    _args: dict[str, Any],
) -> GraphLayout:
    """Run method definition."""
    node_clusters = []
    node_sizes = []

    nodes = list(graph.nodes)

    for node_id in nodes:
        node = graph.nodes[node_id]
        cluster = node.get("cluster", node.get("community", -1))
        node_clusters.append(cluster)
        size = node.get("degree", node.get("size", 0))
        node_sizes.append(size)

    additional_args = {}
    if len(node_clusters) > 0:
        additional_args["node_categories"] = node_clusters
    if len(node_sizes) > 0:
        additional_args["node_sizes"] = node_sizes

    try:
        return get_zero_positions(node_labels=nodes, **additional_args)
    except Exception as e:
        print("Error running zero-position")
        # Umap may fail due to input sparseness or memory pressure.
        # For now, in these cases, we'll just return a layout with all nodes at (0, 0)
        result = []
        for i in range(len(nodes)):
            cluster = node_clusters[i] if len(node_clusters) > 0 else 1
            result.append(
                NodePosition(x=0, y=0, label=nodes[i], size=0, cluster=str(cluster))
            )
        return result


def _run_layout(
    strategy: LayoutGraphStrategyType,
    graphml_or_graph: str | nx.Graph,
    embeddings: NodeEmbeddings,
    args: dict[str, Any],
    # reporter: VerbCallbacks,
) -> GraphLayout:
    graph = load_graph(graphml_or_graph)
    match strategy:
        case LayoutGraphStrategyType.umap:
            return run_umap(
                graph,
                embeddings,
                args,
            )
        case LayoutGraphStrategyType.zero:
            return run_zero(
                graph,
                args,
            )
        case _:
            msg = f"Unknown strategy {strategy}"
            raise ValueError(msg)


def _apply_layout_to_graph(
    graphml_or_graph: str | nx.Graph, layout: GraphLayout
) -> str:
    graph = load_graph(graphml_or_graph)
    for node_position in layout:
        if node_position.label in graph.nodes:
            graph.nodes[node_position.label]["x"] = node_position.x
            graph.nodes[node_position.label]["y"] = node_position.y
            graph.nodes[node_position.label]["size"] = node_position.size
    return "\n".join(nx.generate_graphml(graph))



# Copyright (c) 2024 Microsoft Corporation.
# Licensed under the MIT License

"""A module containing text_embed, load_strategy and create_row_from_embedding_data methods definition."""
import math

import logging
from enum import Enum
from typing import Any, Callable, Dict

import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from . import defs
from text.text_splitter import TokenTextSplitter
from llm.send import EmbLLM, ModelArgs

log = logging.getLogger(__name__)

# Per Azure OpenAI Limits
# https://learn.microsoft.com/en-us/azure/ai-services/openai/reference
DEFAULT_EMBEDDING_BATCH_SIZE = 500

TextEmbedder = Callable[[str], list[float]]


@dataclass
class VectorStoreDocument:
    """A document that is stored in vector storage."""

    id: str | int
    """unique id for the document"""

    text: str | None
    vector: list[float] | None

    attributes: dict[str, Any] = field(default_factory=dict)
    """store any additional metadata, e.g. title, date ranges, etc"""


@dataclass
class VectorStoreSearchResult:
    """A vector storage search result."""

    document: VectorStoreDocument
    """Document that was found."""

    score: float
    """Similarity score between 0 and 1. Higher is more similar."""


class BaseVectorStore(ABC):
    """The base class for vector storage data-access classes."""

    def __init__(
        self,
        collection_name: str,
        db_connection: Any | None = None,
        document_collection: Any | None = None,
        query_filter: Any | None = None,
        **kwargs: Any,
    ):
        self.collection_name = collection_name
        self.db_connection = db_connection
        self.document_collection = document_collection
        self.query_filter = query_filter
        self.kwargs = kwargs

    @abstractmethod
    def connect(self, **kwargs: Any) -> None:
        """Connect to vector storage."""

    @abstractmethod
    def load_documents(
        self, documents: list[VectorStoreDocument], overwrite: bool = True
    ) -> None:
        """Load documents into the vector-store."""

    @abstractmethod
    def similarity_search_by_vector(
        self, query_embedding: list[float], k: int = 10, **kwargs: Any
    ) -> list[VectorStoreSearchResult]:
        """Perform ANN search by vector."""

    @abstractmethod
    def similarity_search_by_text(
        self, text: str, text_embedder: TextEmbedder, k: int = 10, **kwargs: Any
    ) -> list[VectorStoreSearchResult]:
        """Perform ANN search by text."""

    @abstractmethod
    def filter_by_id(self, include_ids: list[str] | list[int]) -> Any:
        """Build a query filter to filter documents by id."""


class TextEmbedStrategyType(str, Enum):
    """TextEmbedStrategyType class definition."""

    openai = "openai"
    mock = "mock"

    def __repr__(self):
        """Get a string representation."""
        return f'"{self.value}"'


def text_embed(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    # cache: PipelineCache,
    send_to: EmbLLM,
    column: str,
    strategy: dict,
    to: str | None = None,
    llm_args: Dict = {},
) -> pd.DataFrame:

    if to is None:
        to = f"{column}_embedding"

    return _text_embed_in_memory(
        input,
        # callbacks,
        # cache,
        send_to,
        column,
        strategy,
        to,
        llm_args,
    )


def _text_embed_in_memory(
    input: pd.DataFrame,
    # callbacks: VerbCallbacks,
    # cache: PipelineCache,
    send_to: EmbLLM,
    column: str,
    strategy: dict,
    to: str,
    llm_args: ModelArgs,
):
    output_df = input
    strategy_args = {**strategy}
    input_table = input

    texts: list[str] = input_table[column].to_numpy().tolist()
    result = run_ai_embed(send_to, texts, strategy_args, llm_args)

    output_df[to] = result.embeddings
    return output_df


@dataclass
class TextEmbeddingResult:
    """Text embedding result class definition."""

    embeddings: list[list[float] | None] | None


def is_null(value: Any) -> bool:
    """Check if value is null or is nan."""

    def is_none() -> bool:
        return value is None

    def is_nan() -> bool:
        return isinstance(value, float) and math.isnan(value)

    return is_none() or is_nan()


def run_ai_embed(
    send_to: EmbLLM,
    input: list[str],
    # callbacks: VerbCallbacks,
    # cache: PipelineCache,
    strategy_args: dict[str, Any],
    llm_args: ModelArgs,
) -> TextEmbeddingResult:
    """Run the Claim extraction chain."""
    if is_null(input):
        return TextEmbeddingResult(embeddings=None)

    batch_size = strategy_args.get("batch_size", 16)
    batch_max_tokens = strategy_args.get("batch_max_tokens", 8191)
    splitter = _get_splitter(batch_max_tokens)

    # Break up the input texts. The sizes here indicate how many snippets are in each input text
    texts, input_sizes = _prepare_embed_texts(input, splitter)
    text_batches = _create_text_batches(
        texts,
        batch_size,
        batch_max_tokens,
        splitter,
    )
    log.info(
        "embedding %d inputs via %d snippets using %d batches. max_batch_size=%d, max_tokens=%d",
        len(input),
        len(texts),
        len(text_batches),
        batch_size,
        batch_max_tokens,
    )
    # ticker = progress_ticker(callbacks.progress, len(text_batches))

    # Embed each chunk of snippets
    embeddings = _execute(send_to, text_batches, llm_args)  # ticker, semaphore)
    embeddings = _reconstitute_embeddings(embeddings, input_sizes)

    return TextEmbeddingResult(embeddings=embeddings)


def _get_splitter(batch_max_tokens: int) -> TokenTextSplitter:
    return TokenTextSplitter(
        encoding_name=defs.ENCODING_MODEL,
        chunk_size=batch_max_tokens,
    )


def _execute(
    send_to: EmbLLM,
    chunks: list[list[str]],
    # tick: ProgressTicker,
    # semaphore: asyncio.Semaphore,
    llm_args: ModelArgs,
) -> list[list[float]]:
    def embed(chunk: list[str]):
        chunk_embeddings = send_to(chunk, llm_args)
        result = np.array(chunk_embeddings)  # .output)
        return result

    results = [embed(chunk) for chunk in chunks]
    # merge results in a single list of lists (reduce the collect dimension)
    return [item for sublist in results for item in sublist]


def _create_text_batches(
    texts: list[str],
    max_batch_size: int,
    max_batch_tokens: int,
    splitter,
) -> list[list[str]]:
    """Create batches of texts to embed."""
    # https://learn.microsoft.com/en-us/azure/ai-services/openai/reference
    # According to this embeddings reference, Azure limits us to 16 concurrent embeddings and 8191 tokens per request
    result = []
    current_batch = []
    current_batch_tokens = 0

    for text in texts:
        token_count = splitter.num_tokens(text)
        if (
            len(current_batch) >= max_batch_size
            or current_batch_tokens + token_count > max_batch_tokens
        ):
            result.append(current_batch)
            current_batch = []
            current_batch_tokens = 0

        current_batch.append(text)
        current_batch_tokens += token_count

    if len(current_batch) > 0:
        result.append(current_batch)

    return result


def _prepare_embed_texts(
    input: list[str],
    splitter,
) -> tuple[list[str], list[int]]:
    sizes: list[int] = []
    snippets: list[str] = []

    for text in input:
        # Split the input text and filter out any empty content
        split_texts = splitter.split_text(text)
        if split_texts is None:
            continue
        split_texts = [text for text in split_texts if len(text) > 0]

        sizes.append(len(split_texts))
        snippets.extend(split_texts)

    return snippets, sizes


def _reconstitute_embeddings(
    raw_embeddings: list[list[float]], sizes: list[int]
) -> list[list[float] | None]:
    """Reconstitute the embeddings into the original input texts."""
    embeddings: list[list[float] | None] = []
    cursor = 0
    for size in sizes:
        if size == 0:
            embeddings.append(None)
        elif size == 1:
            embedding = raw_embeddings[cursor]
            embeddings.append(embedding)
            cursor += 1
        else:
            chunk = raw_embeddings[cursor : cursor + size]
            average = np.average(chunk, axis=0)
            normalized = average / np.linalg.norm(average)
            embeddings.append(normalized.tolist())
            cursor += size
    return embeddings



from enum import Enum
from typing import Any, Callable
import pandas as pd
from pandas.api.types import is_bool
from functools import partial


class MergeStrategy(str, Enum):
    """Merge strategy for merge verb."""

    FirstOneWins = "first one wins"
    LastOneWins = "last one wins"
    Concat = "concat"
    CreateArray = "array"


def _correct_type(value: Any) -> str | int | Any:
    if is_bool(value):
        return str(value).lower()
    try:
        return int(value) if value.is_integer() else value
    except AttributeError:
        return value


def _create_array(column: pd.Series, delim: str) -> str:
    col: pd.DataFrame | pd.Series = column.dropna().apply(lambda x: _correct_type(x))
    return delim.join(col.astype(str))


merge_strategies: dict[MergeStrategy, Callable] = {
    MergeStrategy.FirstOneWins: lambda values, **_kwargs: values.dropna().apply(
        lambda x: _correct_type(x)
    )[0],
    MergeStrategy.LastOneWins: lambda values, **_kwargs: values.dropna().apply(
        lambda x: _correct_type(x)
    )[-1],
    MergeStrategy.Concat: lambda values, delim, **_kwargs: _create_array(values, delim),
    MergeStrategy.CreateArray: lambda values, **_kwargs: _create_array(values, ","),
}


def merge(
    table: pd.DataFrame,
    to: str,
    columns: list[str],
    strategy: str,
    delimiter: str = "",
    preserveSource: bool = False,  # noqa: N803
) -> pd.DataFrame:
    """Merge verb implementation."""
    merge_strategy = MergeStrategy(strategy)

    table[to] = table[columns].apply(
        partial(merge_strategies[merge_strategy], delim=delimiter), axis=1
    )

    if not preserveSource:
        table.drop(columns=columns, inplace=True)

    return table



import pandas as pd
import df_ops
import os
from utils.save import parquet_table_load, parquet_table_save


def create_base_documents(
    final_text_units_output: pd.DataFrame,
    query_output_dir: str | None = None,
    save: bool = True,
    try_load: bool = True,
):

    fn_name = "create_base_documents"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)

    dataset = df_ops.unroll(final_text_units_output, **{"column": "document_ids"})
    dataset = df_ops.select(dataset, columns=["id", "document_ids", "text"])
    rename_chunk_doc_id = df_ops.rename(
        dataset,
        **{
            "columns": {
                "document_ids": "chunk_doc_id",
                "id": "chunk_id",
                "text": "chunk_text",
            }
        },
    )
    dataset = df_ops.join(
        rename_chunk_doc_id,
        other=final_text_units_output,  # ?
        **{
            # Join the doc id from the chunk onto the original document
            "on": ["chunk_doc_id", "id"]
        },
    )

    docs_with_text_units = df_ops.aggregate_override(
        dataset,
        **{
            "groupby": ["id"],
            "aggregations": [
                {
                    "column": "chunk_id",
                    "operation": "array_agg",
                    "to": "text_units",
                }
            ],
        },
    )

    dataset = df_ops.join(
        docs_with_text_units,
        final_text_units_output,
        **{
            "on": ["id", "id"],
            "strategy": "right outer",
        },
    )

    dataset = df_ops.rename(dataset, **{"columns": {"text": "raw_content"}})

    base_documents_output = df_ops.convert(
        dataset, **{"column": "id", "to": "id", "type": "string"}
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, base_documents_output)

    return base_documents_output



import os
import df_ops
import pandas as pd

from utils.save import parquet_table_load, parquet_table_save


def create_final_relationships(
    base_entity_graph_output: pd.DataFrame,
    final_nodes_output: pd.DataFrame,
    query_output_dir: str,
    save: bool = True,
    try_load: bool = True,
):
    fn_name = "create_final_relationships"
    if try_load and os.path.exists(
        os.path.join(query_output_dir, f"{fn_name}.parquet")
    ):
        return parquet_table_load(query_output_dir, fn_name)

    dataset = df_ops.unpack_graph(
        base_entity_graph_output,
        **{
            "column": "clustered_graph",
            "type": "edges",
        },
    )

    dataset = df_ops.rename(
        dataset,
        **{"columns": {"source_id": "text_unit_ids"}},
    )

    dataset = df_ops.filter_verb(
        dataset,
        column="level",
        strategy="value",
        operator="equals",
        value=0,
    )

    pruned_edges = df_ops.drop(
        dataset,
        columns=["level"],
    )

    filtered_nodes = df_ops.filter_verb(
        final_nodes_output,
        column="level",
        strategy="value",
        operator="equals",
        value=0,
    )

    dataset = df_ops.compute_edge_combined_degree(
        input=pruned_edges,
        nodes=filtered_nodes,
        **{"to": "rank"},
    )

    dataset = df_ops.convert(
        dataset,
        **{
            "column": "human_readable_id",
            "type": "string",
            "to": "human_readable_id",
        },
    )

    dataset = df_ops.convert(
        dataset,
        **{
            "column": "text_unit_ids",
            "type": "array",
            "to": "text_unit_ids",
        },
    )

    if save:
        parquet_table_save(query_output_dir, fn_name, dataset)

    return dataset



import pandas as pd
from typing import Any

def drop(table: pd.DataFrame, columns: list[str], **_kwargs: Any) -> pd.DataFrame:
    """Drop verb implementation."""
    return table.drop(columns=columns)